#
# Abbas Razaghpanah (arazaghpanah@cs.stonybrook.edu)
# November 2016, Stony Brook University
#
# tcp_connect.py: TCP connect test

from datetime import datetime
import logging
import socket
import time
import threading

MAX_THREAD_START_RETRY = 10
THREAD_START_DELAY = 3

def tcp_connect(host, port, external=None, log_prefix=''):
    result = {
        "host" : host,
        "port" : port
    }

    logging.debug(log_prefix + "Testing TCP connect to %s:%s..." % (host, port))

    try:
        ip = socket.gethostbyname(host)
        if (ip is not None):
            result["ip"] = ip
    except Exception as err:
        result["ip_err"] = str(err)

    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setblocking(True)
        sock.settimeout(5)
        start_time = datetime.now()
        sock.connect((host, int(port)))
        end_time = datetime.now()
        elapsed = int((end_time - start_time).total_seconds() * 1000)
        sock.close()
        result["success"] = "true"
    except Exception as err:
        end_time = datetime.now()
        elapsed = int((end_time - start_time).total_seconds() * 1000)
        result["failure"] = str(err)

    result["time"] = str(elapsed)

    # the external result is used when threading to store
    # the results in the list container provided.
    if external is not None and type(external) is dict:
        external[host + ":" + str(port)] = result

    return result

def tcp_connect_batch(input_list, results={}, delay_time=0.1, max_threads=100):
    """
    This is a parallel version of the TCP connect primitive.

    :param input_list: the input is a list of host/port pairs
    :param delay_time: delay before starting each thread
    :param max_threads: maximum number of concurrent threads
    :return:
    """
    threads = []
    thread_error = False
    thread_wait_timeout = 200
    ind = 1
    total_item_count = len(input_list)
    for host,port in input_list:
        wait_time = 0
        while threading.active_count() > max_threads:
            time.sleep(1)
            wait_time += 1
            if wait_time > thread_wait_timeout:
                thread_error = True
                break

        if thread_error:
            results["error"] = "Threads took too long to finish."
            break

        # add just a little bit of delay before starting the thread
        # to avoid overwhelming the connection.
        time.sleep(delay_time)

        log_prefix = "%d/%d: " % (ind, total_item_count)
        thread = threading.Thread(target=tcp_connect,
                                  args=(host, port,
                                        results, log_prefix))
        ind += 1
        thread.setDaemon(1)
        thread_open_success = False
        retries = 0
        while not thread_open_success and retries < MAX_THREAD_START_RETRY:
            try:
                thread.start()
                threads.append(thread)
                thread_open_success = True
            except:
                retries += 1
                time.sleep(THREAD_START_DELAY)
                logging.error("%sThread start failed for %s, retrying... (%d/%d)" % (log_prefix, host, retries, MAX_THREAD_START_RETRY))

        if retries == MAX_THREAD_START_RETRY:
            logging.error("%sCan't start a new thread for %s after %d retries." % (log_prefix, host, retries))

    for thread in threads:
        thread.join(thread_wait_timeout)

    return results



import bz2
import glob
import imp
import json
import logging
import logging.config
import os
import signal
import sys
import tarfile
import time
from datetime import datetime

import centinel
from centinel.backend import get_meta
from centinel.primitives.tcpdump import Tcpdump
from experiment import ExperimentList
from centinel.vpn.cli import get_external_ip

loaded_modules = set()
# we need a global reference to stop it if we receive an interrupt.
tds = []

def signal_handler(signal, frame):
        logging.warn('Interrupt signal received.')
        if len(tds) > 0:
            logging.warn('Stopping TCP dump...')
            for td in tds:
                td.stop()
                td.delete()
        logging.warn('Exiting...')
        sys.exit(0)

signal.signal(signal.SIGTERM, signal_handler)

class Client:
    def __init__(self, config, vpn_provider=None):
        self.config = config
        self.experiments = self.load_experiments()
        self._meta = None
        self.vpn_provider = vpn_provider

    def setup_logging(self):

        log_config = {'version': 1,
                      'formatters': {'error': {'format': self.config['log']['log_format']},
                                     'debug': {'format': self.config['log']['log_format']}},
                      'handlers': {'console': {'class': 'logging.StreamHandler',
                                               'formatter': 'debug',
                                               'level': self.config['log']['log_level']},
                                   'file': {'class': 'logging.FileHandler',
                                            'filename': self.config['log']['log_file'],
                                            'formatter': 'error',
                                            'level': self.config['log']['log_level']}},
                      'root': {'handlers': ('console', 'file'), 'level': self.config['log']['log_level']}}
        logging.config.dictConfig(log_config)

        logging.debug("Finished setting up logging.")

    def get_result_file(self, name, start_time):
        result_file = "%s-%s.json.bz2" % (name, start_time)
        return os.path.join(self.config['dirs']['results_dir'], result_file)

    def get_input_file(self, experiment_name):
        input_file = "%s" % experiment_name
        return os.path.join(self.config['dirs']['data_dir'], input_file)

    def load_input_file(self, name):
        input_file = self.get_input_file(name)

        if not os.path.isfile(input_file):
            logging.error("Input file not found %s" % input_file)
            return None

        try:
            input_file_handle = open(input_file)
        except Exception as exp:
            logging.exception("Can not read from %s: %s" % (input_file, str(exp)))
            return None
        logging.debug("Input file %s loaded." % name)
        return input_file_handle

    def load_experiments(self):
        """This function will return the list of experiments.
        """
        logging.debug("Loading experiments.")
        # look for experiments in experiments directory
        exp_dir = self.config['dirs']['experiments_dir']
        for path in glob.glob(os.path.join(exp_dir, '[!_]*.py')):
            # get name of file and path
            name, ext = os.path.splitext(os.path.basename(path))
            # load the experiment
            try:
                # do not load modules that have already been loaded
                if name in loaded_modules:
                    continue
                imp.load_source(name, path)
                loaded_modules.add(name)
                logging.debug("Loaded experiment \"%s(%s)\"." % (name, path))
            except Exception as exception:
                logging.exception("Failed to load experiment %s: %s" %
                                  (name, exception))

        logging.debug("Finished loading experiments.")
        # return dict of experiment names and classes
        return ExperimentList.experiments

    def has_experiments_to_run(self):
        # load scheduler information
        sched_filename = os.path.join(self.config['dirs']['experiments_dir'],
                                      'scheduler.info')
        sched_info = {}
        if os.path.exists(sched_filename):
            with open(sched_filename, 'r') as file_p:
                sched_info = json.load(file_p)

        for name in sched_info:
            run_next = sched_info[name]['last_run']
            run_next += sched_info[name]['frequency']
            if run_next <= time.time():
                logging.debug("Client has experiment(s) to run (%s)." % name)
                return True
        logging.debug("Client has no experiments to run.")
        return False

    def get_meta(self):
        """we only want to get the meta information (our normalized IP) once,
        so we are going to do lazy instantiation to improve performance

        """
        # get the normalized IP if we don't already have it
        if self._meta is None:
            external_ip = get_external_ip()
            if external_ip:
                self._meta = get_meta(self.config, external_ip)
            else:
                raise Exception("Unable to get public IP")
            if 'custom_meta' in self.config:
                self._meta['custom_meta'] = self.config['custom_meta']

        return self._meta

    def run(self, data_dir=None):

        """
        Note: this function will check the experiments directory for a
        special file, scheduler.info, that details how often each
        experiment should be run and the last time the experiment was
        run. If the time since the experiment was run is shorter than
        the scheduled interval in seconds, then the experiment will
        not be run.

        :param data_dir:
        :return:
        """
        # XXX: android build needs this. refactor
        if data_dir:
            centinel_home = data_dir
            self.config['dirs']['results_dir'] = os.path.join(centinel_home,
                                                              'results')

        logging.info('Centinel started.')

        if not os.path.exists(self.config['dirs']['results_dir']):
            logging.warn("Creating results directory in "
                         "%s" % (self.config['dirs']['results_dir']))
            os.makedirs(self.config['dirs']['results_dir'])
        logging.debug("Results directory: %s" % (self.config['dirs']['results_dir']))

        # load scheduler information
        sched_filename = os.path.join(self.config['dirs']['experiments_dir'],
                                      'scheduler.info')

        logging.debug("Loading scheduler file.")
        sched_info = {}
        if os.path.exists(sched_filename):
            with open(sched_filename, 'r') as file_p:
                try:
                    sched_info = json.load(file_p)
                except Exception as exp:
                    logging.error("Failed to load the "
                                  "scheduler: %s" % str(exp))
                    return

        logging.debug("Scheduler file loaded.")

        logging.debug("Processing the experiment schedule.")
        for name in sched_info:

            # check if we should preempt on the experiment (if the
            # time to run next is greater than the current time) and
            # store the last run time as now
            #
            # Note: if the experiment is not in the scheduler, then it
            # will not be run at all.
            run_next = sched_info[name]['last_run']
            run_next += sched_info[name]['frequency']
            if run_next > time.time():
                run_next_str = datetime.fromtimestamp(long(run_next))
                logging.debug("Skipping %s, it will "
                              "be run on or after %s." % (name, run_next_str))
                continue

            # backward compatibility with older-style scheduler
            if 'python_exps' not in sched_info[name]:
                self.run_exp(name=name)
            else:
                exps = sched_info[name]['python_exps'].items()
                for python_exp, exp_config in exps:
                    logging.debug("Running %s." % python_exp)
                    self.run_exp(name=python_exp, exp_config=exp_config, schedule_name=name)
                    logging.debug("Finished running %s." % python_exp)
            sched_info[name]['last_run'] = time.time()

        logging.debug("Updating timeout values in scheduler.")
        # write out the updated last run times
        with open(sched_filename, 'w') as file_p:
            json.dump(sched_info, file_p, indent=2,
                      separators=(',', ': '))

        self.consolidate_results()

        logging.info("Finished running experiments. "
                     "Look in %s for results." % (self.config['dirs']['results_dir']))

    def run_exp(self, name, exp_config=None, schedule_name=None):
        if name[-3:] == ".py":
            name = name[:-3]
        if name not in self.experiments:
            logging.error("Experiment file %s not found! Skipping." % name)
        else:
            exp_class = self.experiments[name]
            results = {"meta": {}}
            try:
                logging.debug("Getting metadata for experiment...")
                meta = self.get_meta()
                results["meta"] = meta
            except Exception as exception:
                logging.exception("Error fetching metadata for "
                                  "%s: %s" % (name, exception))
                results["meta_exception"] = str(exception)

            if schedule_name is not None:
                results["meta"]["schedule_name"] = schedule_name
            else:
                results["meta"]["schedule_name"] = name

            start_time = datetime.now()
            results["meta"]["client_time"] = start_time.isoformat()

            results["meta"]["centinel_version"] = centinel.__version__

            # include vpn provider in metadata
            if self.vpn_provider:
                results["meta"]["vpn_provider"] = self.vpn_provider

            input_files = {}
            if exp_config is not None:
                if (('input_files' in exp_config) and
                        (exp_config['input_files'] is not None)):
                    for filename in exp_config['input_files']:
                        file_handle = self.load_input_file(filename)
                        if file_handle is not None:
                            input_files[filename] = file_handle

                if (('params' in exp_config) and
                        (exp_config['params'] is not None)):
                    exp_class.params = exp_config['params']

            # if the scheduler does not specify input files, but
            # the experiment class specifies a list of input file names,
            # load them. failing to load input files does not stop
            # experiment from running.
            if len(input_files) == 0:
                if exp_class.input_files is not None:
                    for filename in exp_class.input_files:
                        file_handle = self.load_input_file(filename)
                        if file_handle is not None:
                            input_files[filename] = file_handle
                # otherwise, fall back to [schedule name].txt (deprecated)
                else:
                    filename = "%s.txt" % name
                    file_handle = self.load_input_file(filename)
                    if file_handle is not None:
                        input_files[filename] = file_handle

            try:
                # instantiate the experiment
                logging.debug("Initializing the experiment class for %s" % name)

                # these constants can be useful for some experiments, but it is not
                # encouraged to use these directly
                global_constants = {'experiments_dir': self.config['dirs']['experiments_dir'],
                                    'results_dir': self.config['dirs']['results_dir'],
                                    'data_dir': self.config['dirs']['data_dir']}

                exp_class.global_constants = global_constants

                exp = exp_class(input_files)
            except Exception as exception:
                logging.exception("Error initializing %s: %s" % (name, exception))
                results["init_exception"] = str(exception)
                return

            exp.global_constants = global_constants

            run_tcpdump = True

            if self.config['results']['record_pcaps'] is False:
                logging.info("Your configuration has disabled pcap "
                             "recording, tcpdump will not start.")
                run_tcpdump = False
                # disable this on the experiment too
                exp.record_pcaps = False

            if run_tcpdump and os.geteuid() != 0:
                logging.info("Centinel is not running as root, "
                             "tcpdump will not start.")
                run_tcpdump = False

            if run_tcpdump and exp_class.overrides_tcpdump:
                logging.info("Experiment overrides tcpdump recording.")
                run_tcpdump = False

            tcpdump_started = False

            try:
                if run_tcpdump:
                    td = Tcpdump()
                    tds.append(td)
                    td.start()
                    tcpdump_started = True
                    logging.info("tcpdump started...")
                    # wait for tcpdump to initialize
                    time.sleep(2)
            except Exception as exp:
                logging.exception("Failed to run tcpdump: %s" % (exp,))

            try:
                # run the experiment
                exp.run()
            except Exception as exception:
                logging.exception("Error running %s: %s" % (name, exception))
                results["runtime_exception"] = str(exception)
            except KeyboardInterrupt:
                logging.warn("Keyboard interrupt received, stopping experiment...")


            # save any external results that the experiment has generated
            # they could be anything that doesn't belong in the json file
            # (e.g. pcap files)
            # these should all be compressed with bzip2
            # the experiment is responsible for giving these a name and
            # keeping a list of files in the json results
            results_dir = self.config['dirs']['results_dir']
            if exp.external_results is not None:
                logging.debug("Writing external files for %s" % name)
                for fname, fcontents in exp.external_results.items():
                    external_file_name = ("external_%s-%s-%s"
                                          ".bz2" % (name,
                                                    start_time.strftime("%Y-%m-%dT%H%M%S.%f"),
                                                    fname))
                    external_file_path = os.path.join(results_dir,
                                                      external_file_name)
                    try:
                        with open(external_file_path, 'w:bz2') as file_p:
                            data = bz2.compress(fcontents)
                            file_p.write(data)
                            logging.debug("External file "
                                          "%s written successfully" % fname)
                    except Exception as exp:
                        logging.exception("Failed to write external file:"
                                          "%s" % exp)
                logging.debug("Finished writing external files for %s" % name)

            if tcpdump_started:
                logging.info("Waiting for tcpdump to process packets...")
                # 5 seconds should be enough. this hasn't been tested on
                # a RaspberryPi or a Hummingboard i2
                time.sleep(5)
                td.stop()
                logging.info("tcpdump stopped.")
                bz2_successful = False
                data = None
                try:
                    pcap_file_name = ("pcap_%s-%s.pcap"
                                      ".bz2" % (name, start_time.strftime("%Y-%m-%dT%H%M%S.%f")))
                    pcap_file_path = os.path.join(results_dir,
                                                  pcap_file_name)

                    with open(pcap_file_path, 'wb') as pcap_bz2, open(td.pcap_filename(), 'rb') as pcap:
                        compressor = bz2.BZ2Compressor()
                        compressed_size_so_far = 0
                        for pcap_data in iter(lambda: pcap.read(10 * 1024), b''):
                            compressed_chunk  = compressor.compress(pcap_data)
                            pcap_bz2.write(compressed_chunk)

                            if len(compressed_chunk):
                                compressed_size_so_far += len(compressed_chunk)

                        compressed_chunk = compressor.flush()
                        pcap_bz2.write(compressed_chunk)

                        if len(compressed_chunk):
                            compressed_size_so_far += len(compressed_chunk)
                        uncompressed_size = os.path.getsize(td.pcap_filename())
                        compression_ratio = 100 * (float(compressed_size_so_far) / float(uncompressed_size))
                        logging.debug("pcap BZ2 compression: compressed/uncompressed (ratio):"
                                      " %d/%d (%.1f%%)" % (compressed_size_so_far, uncompressed_size, compression_ratio))

                    logging.info("Saved pcap to "
                                 "%s." % pcap_file_path)
                    bz2_successful = True
                except Exception as exception:
                    logging.exception("Failed to compress and write "
                                      "pcap file: %s" % exception)
                if not bz2_successful:
                    logging.info("Writing pcap file uncompressed")
                    try:
                        pcap_file_name = ("pcap_%s-%s"
                                          ".pcap" % (name, start_time.strftime("%Y-%m-%dT%H%M%S.%f")))
                        pcap_file_path = os.path.join(results_dir,
                                                      pcap_file_name)

                        with open(pcap_file_path, 'wb') as pcap_out, open(td.pcap_filename(), 'rb') as pcap:
                            for pcap_data in iter(lambda: pcap.read(10 * 1024), b''):
                                pcap_out.write(pcap_data)

                        logging.info("Saved pcap to "
                                     "%s." % pcap_file_path)
                    except Exception as exception:
                        logging.exception("Failed to write "
                                          "pcap file: %s" % exception)
                # delete pcap data to free up some memory
                logging.debug("Removing pcap data from memory")
                td.delete()
                del data
                del td

            # close input file handle(s)
            logging.debug("Closing input files for %s" % name)
            if type(input_files) is dict:
                for file_name, file_handle in input_files.items():
                    try:
                        file_handle.close()
                    except AttributeError:
                        logging.warning("Closing %s failed" % file_name)
            logging.debug("Input files closed for %s" % name)

            logging.debug("Storing results for %s" % name)
            try:
                results[name] = exp.results
            except Exception as exception:
                logging.exception("Error storing results for "
                                  "%s: %s" % (name, exception))
                if "results_exception" not in results:
                    results["results_exception"] = {}

                results["results_exception"][name] = str(exception)

            end_time = datetime.now()
            time_taken = (end_time - start_time)
            results["meta"]["time_taken"] = time_taken.total_seconds()

            logging.info("%s took %s to finish." % (name, time_taken))

            logging.debug("Saving %s results to file" % name)
            try:
                # pretty printing results will increase file size, but files are
                # compressed before sending.
                result_file_path = self\
                    .get_result_file(name, start_time.strftime("%Y-%m-%dT%H%M%S.%f"))
                result_file = bz2.BZ2File(result_file_path, "w")
                json.dump(results, result_file, indent=2, separators=(',', ': '),
                    # ignore encoding errors, these will be dealt with on the server
                    ensure_ascii=False)
                result_file.close()

                # free up memory by deleting results from memory
                del results
                del result_file
            except Exception as exception:
                logging.exception("Error saving results for "
                                  "%s to file: %s" % (name, exception))
            logging.debug("Done saving %s results to file" % name)

    def consolidate_results(self):
        # bundle and compress result files
        result_files = [path for path in glob.glob(
            os.path.join(self.config['dirs']['results_dir'], '*.json.bz2'))]

        if len(result_files) >= self.config['results']['files_per_archive']:
            logging.info("Compressing and archiving results.")

            files_archived = 0
            archive_count = 0
            tar_file = None
            files_per_archive = self.config['results']['files_per_archive']
            results_dir = self.config['dirs']['results_dir']
            for path in result_files:
                if (files_archived % files_per_archive) == 0:
                    archive_count += 1
                    archive_filename = "results-%s_%d.tar.bz2" % (
                        datetime.now().strftime("%Y-%m-%dT%H%M%S.%f"), archive_count)
                    archive_file_path = os.path.join(results_dir,
                                                     archive_filename)
                    logging.info("Creating new archive"
                                 " %s" % archive_file_path)
                    if tar_file:
                        tar_file.close()
                    tar_file = tarfile.open(archive_file_path, "w:bz2")

                tar_file.add(path, arcname=os.path.basename(path))
                os.remove(path)
                files_archived += 1

            if tar_file:
                tar_file.close()



import getpass
import json
import logging
import os

import centinel


class Configuration:

    def __init__(self):

        self.params = {}

        # version info
        version_info = {'version': centinel.__version__}
        self.params['version'] = version_info

        # centinel user
        user_info = {'current_user': getpass.getuser()}
        user_home = os.path.expanduser('~' + user_info['current_user'])
        user_info['centinel_home'] = os.path.join(user_home, '.centinel')
        user_info['is_vpn'] = False
        self.params['user'] = user_info

        # directory structure
        # by default, *ALWAYS* use the directories in the home directory
        dirs = {
            'experiments_dir': os.path.join(self.params['user']['centinel_home'], 'experiments'),
            'data_dir': os.path.join(self.params['user']['centinel_home'], 'data'),
            'results_dir': os.path.join(self.params['user']['centinel_home'], 'results')
        }
        self.params['dirs'] = dirs

        # if we creating a new config file, then we can expect to be
        # setting up the the home directory, so we should create all
        # of these directories if they don't exist
        for dir_key in self.params['dirs']:
            directory = self.params['dirs'][dir_key]
            if not os.path.exists(directory):
                os.makedirs(directory)

        # results
        results = {'delete_after_sync': True,
                   'files_per_archive': 10,
                   'record_pcaps': True,
                   'upload_pcaps': True}
        self.params['results'] = results

        # logging
        self.params['log'] = {}
        self.params['log']['log_level'] = logging.INFO
        self.params['log']['log_file'] = os.path.join(self.params['user']['centinel_home'],
                                                      "centinel.log")
        self.params['log']['log_format'] = '%(asctime)s %(filename)s(line %(lineno)d) ' \
                                           '%(levelname)s: %(message)s'

        # experiments
        experiments = {'tcpdump_params': ["-i", "any"]}
        self.params['experiments'] = experiments

        # server
        servers = {'server_url': "https://server.iclab.cs.stonybrook.edu:8082",
                   'login_file': os.path.join(self.params['user']['centinel_home'], 'login'),
                   # the entire transaction should take less than 30 min
                   'total_timeout': 60*30,
                   # set a socket timeout of 15 seconds (no way to do per request
                   # platform independently)
                   'req_timeout': 15,
                   'verify': True}
        self.params['server'] = servers

        # proxy
        proxy = {'proxy_type': None,    # "socks" or "http"
                 'proxy_url': None,     # "http://127.0.0.1:9050"
                 'proxy': None}
        # TODO: this if clause looks useless to me (Adrian)
        if proxy['proxy_type']:
            proxy['proxy'] = {proxy['proxy_type']: proxy['proxy_url']}
        self.params['proxy'] = proxy

    def parse_config(self, config_file):
        """
        Given a configuration file, read in and interpret the results

        :param config_file:
        :return:
        """

        with open(config_file, 'r') as f:
            config = json.load(f)
        self.params = config
        if self.params['proxy']['proxy_type']:
            self.params['proxy'] = {self.params['proxy']['proxy_type']:
                                    self.params['proxy']['proxy_url']}

    def update(self, old, backup_path=None):
        """
        Update the old configuration file with new values.

        :param old: old configuration to update.
        :param backup_path: path to write a backup of the old config file.

        :return:
        """
        for category in old.params.keys():
            for parameter in old.params[category].keys():
                if (category in self.params and parameter in self.params[category] and
                        (old.params[category][parameter] != self.params[category][parameter]) and
                        (category != "version")):
                    print ("Config value '%s.%s' "
                           "in old configuration is different "
                           "from the new version\n"
                           "[old value] = %s\n"
                           "[new value] = %s"
                           "" % (category, parameter,
                                 old.params[category][parameter],
                                 self.params[category][parameter]))
                    answer = raw_input("Do you want to overwrite? ([y]/n) ")
                    while answer.lower() not in ['y', 'yes', 'n', 'no']:
                        answer = raw_input("Answer not recongnized. Enter 'y' or 'n'. ")

                    if answer in ['n', 'no']:
                        old_value = old.params[category][parameter]
                        self.params[category][parameter] = old_value
                elif not (category in self.params and
                          parameter in self.params[category]):
                    print ("Deprecated config option '%s.%s' has "
                           "been removed." % (category, parameter))
        if backup_path is not None:
            old.write_out_config(backup_path)
            print "Backup saved in %s." % backup_path

    def write_out_config(self, config_file):
        """
        Write out the configuration file

        :param config_file:
        :return:

        Note: this will erase all comments from the config file

        """
        with open(config_file, 'w') as f:
            json.dump(self.params, f, indent=2,
                      separators=(',', ': '))



#
# Abbas Razaghpanah (arazaghpanah@cs.stonybrook.edu)
# February 2015, Stony Brook University
#
# tracerout.py: interface to system traceroute
# for this to work, traceroute has to be installed
# and accessible.

import copy
import logging
import threading
import time
from sys import platform

import trparse

from centinel import command

MAX_THREAD_START_RETRY = 10
THREAD_START_DELAY = 3

def traceroute(domain, method="udp", cmd_arguments=None,
               external=None, log_prefix=''):
    """
    This function uses centinel.command to issue
    a traceroute command, wait for it to finish execution and
    parse the results out to a dictionary.

    :param domain: the domain to be queried
    :param method: the packet type used for traceroute, UDP by default
    :param cmd_arguments: the list of arguments that need to be passed
                          to traceroute.
    :param external:
    :param log_prefix:
    :return:
    """

    # the method specified by the function parameter here will
    # over-ride the ones given in cmd_arguments because
    # traceroute will use the last one in the argument list.
    _cmd_arguments = []

    logging.debug("%sRunning traceroute for "
                  "%s using %s probes." % (log_prefix, domain, method))

    results = {"method": method}

    if cmd_arguments is not None:
        _cmd_arguments = copy.deepcopy(cmd_arguments)

    if method == "tcp":
        if platform in ['linux', 'linux2']:
            _cmd_arguments.append('-T')
        elif platform == "darwin":
            _cmd_arguments.append('-P')
            _cmd_arguments.append('tcp')

    elif method == "udp":
        if platform in ['linux', 'linux2']:
            _cmd_arguments.append('-U')
        elif platform == "darwin":
            _cmd_arguments.append('-P')
            _cmd_arguments.append('udp')

    elif method == "icmp":
        if platform in ['linux', 'linux2']:
            _cmd_arguments.append('-I')
        elif platform == "darwin":
            _cmd_arguments.append('-P')
            _cmd_arguments.append('icmp')

    cmd = ['traceroute'] + _cmd_arguments + [domain]

    caller = command.Command(cmd, _traceroute_callback)
    caller.start()
    if not caller.started:
        if caller.exception is not None:
            if "No such file or directory" in caller.exception:
                message = "traceroute not found or not installed"
            else:
                message = ("traceroute thread threw an "
                           "exception: %s" % caller.exception)
        elif "enough privileges" in caller.notifications:
            message = "not enough privileges"
        elif "not known" in caller.notifications:
            message = "name or service not known"
        else:
            message = caller.notifications

        results["dest_name"] = domain
        results["error"] = message
        if external is not None and type(external) is dict:
            external[domain] = results
        return results

    forcefully_terminated = False
    timeout = 60
    start_time = time.time()
    # check every second to see if the execution has stopped
    while caller.thread.isAlive():
        if (time.time() - start_time) > timeout:
            caller.stop()
            forcefully_terminated = True
            break
        time.sleep(1)
    # we are only accurate down to seconds, so we have
    # to round up
    time_elapsed = int(time.time() - start_time)

    output_string = caller.notifications

    try:
        parsed_output = trparse.loads(output_string)
    except Exception as exc:
        results["dest_name"] = domain
        results["error"] = str(exc)
        results["raw"] = output_string
        if external is not None and type(external) is dict:
            external[domain] = results
        return results

    hops = list()
    for hop in parsed_output.hops:
        hop_json = {"index": hop.idx, "asn": hop.asn}
        probes_json = []
        for probe in hop.probes:
            probes_json.append({"name": probe.name,
                                "ip": probe.ip,
                                "rtt": probe.rtt,
                                "anno": probe.anno})
        hop_json["probes"] = probes_json
        hops.append(hop_json)

    results["dest_name"] = parsed_output.dest_name
    results["dest_ip"] = parsed_output.dest_ip
    results["hops"] = hops
    results["forcefully_terminated"] = forcefully_terminated
    results["time_elapsed"] = time_elapsed

    # the external result is used when threading to store
    # the results in the list container provided.
    if external is not None and type(external) is dict:
        external[domain] = results

    return results


def traceroute_batch(input_list, results={}, method="udp", cmd_arguments=None,
                     delay_time=0.1, max_threads=100):
    """
    This is a parallel version of the traceroute primitive.

    :param input_list: the input is a list of domain names
    :param method: the packet type used for traceroute, UDP by default
    :param cmd_arguments: the list of arguments that need to be passed
                        to traceroute.
    :param delay_time: delay before starting each thread
    :param max_threads: maximum number of concurrent threads
    :return:
    """
    threads = []
    thread_error = False
    thread_wait_timeout = 200
    ind = 1
    total_item_count = len(input_list)
    for domain in input_list:
        wait_time = 0
        while threading.active_count() > max_threads:
            time.sleep(1)
            wait_time += 1
            if wait_time > thread_wait_timeout:
                thread_error = True
                break

        if thread_error:
            results["error"] = "Threads took too long to finish."
            break

        # add just a little bit of delay before starting the thread
        # to avoid overwhelming the connection.
        time.sleep(delay_time)

        log_prefix = "%d/%d: " % (ind, total_item_count)
        thread = threading.Thread(target=traceroute,
                                  args=(domain, method, cmd_arguments,
                                        results, log_prefix))
        ind += 1
        thread.setDaemon(1)
        thread_open_success = False
        retries = 0
        while not thread_open_success and retries < MAX_THREAD_START_RETRY:
            try:
                thread.start()
                threads.append(thread)
                thread_open_success = True
            except:
                retries += 1
                time.sleep(THREAD_START_DELAY)
                logging.error("%sThread start failed for %s, retrying... (%d/%d)" % (log_prefix, domain, retries, MAX_THREAD_START_RETRY))

        if retries == MAX_THREAD_START_RETRY:
            logging.error("%sCan't start a new thread for %s after %d retries." % (log_prefix, domain, retries))

    for thread in threads:
        thread.join(thread_wait_timeout)

    return results


def _traceroute_callback(self, line, kill_switch):
    """
    Callback function to handle traceroute.

    :param self:
    :param line:
    :param kill_switch:
    :return:
    """

    line = line.lower()
    if "traceroute to" in line:
        self.started = True
    # need to run as root but not running as root.
    # usually happens when doing TCP and ICMP traceroute.
    if "enough privileges" in line:
        self.error = True
        self.kill_switch()
        self.stopped = True
    # name resolution failed
    if "service not known" in line:
        self.error = True
        self.kill_switch()
        self.stopped = True



#!/usr/bin/python
# vpn.py: a top level interface to commands to run VPNs on a VM as
# different clients. The use case is allowing us to measure from more
# places.

import argparse
import logging
from random import shuffle
import os
import time
import sys
import signal
import dns.resolver
import json

import centinel.backend
import centinel.client
import centinel.config
import centinel.vpn.openvpn as openvpn
import centinel.vpn.hma as hma
import centinel.vpn.ipvanish as ipvanish
import centinel.vpn.purevpn as purevpn
import centinel.vpn.vpngate as vpngate

PID_FILE = "/tmp/centinel.lock"


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--auth-file', '-u', dest='auth_file', default=None,
                        help=("File with HMA username on first line, \n"
                              "HMA password on second line"))
    parser.add_argument('--crt-file', '-r', dest='crt_file', default=None,
                        help=("Certificate file for the current vpn\n"
                              "provider, if provided"))
    parser.add_argument('--tls-auth', '-t', dest='tls_auth', default=None,
                        help="Key for additional layer of authentication")
    parser.add_argument('--key-direction', '-k', dest='key_direction', default=None,
                        help=("Key direction for tls auth, must specify when "
                              "tls-auth is used"))
    parser.add_argument('--reduce-endpoint', dest='reduce_vp',
                        action="store_true", default=False,
                        help="Reduce the number of vantage points by only connect to "
                             "one vantage point for each AS and country combination")
    g1 = parser.add_mutually_exclusive_group()
    g1.add_argument('--create-hma-configs', dest='create_HMA',
                    action="store_true",
                    help='Create the openvpn config files for HMA')
    g1.add_argument('--create-ipvanish-configs', dest='create_IPVANISH',
                    action='store_true',
                    help='Create the openvpn config files for IPVanish')
    g1.add_argument('--create-purevpn-configs', dest='create_PUREVPN',
                    action='store_true',
                    help='Create the openvpn config files for PureVPN')
    g1.add_argument('--create-vpngate-configs', dest='create_VPNGATE',
                    action='store_true',
                    help='Create the openvpn config files for VPN Gate')
    parser.add_argument('--shuffle', '-s', dest='shuffle_lists',
                        action="store_true", default=False,
                        help='Randomize the order of vantage points')
    parser.add_argument('--exclude', '-e', dest='exclude_list', default=None,
                        help=('Countries to exclude when scanning (comma '
                              'separated two letter country codes)'))
    parser.add_argument('--log-file', '-l', dest='log_file', default=None,
                        help="Log file location")
    g2 = parser.add_mutually_exclusive_group(required=True)
    g2.add_argument('--directory', '-d', dest='directory',
                    help='Directory with experiments, config files, etc.')
    create_conf_help = ('Create configuration files for the given '
                        'openvpn config files so that we can treat each '
                        'one as a client. The argument should be a '
                        'directory with a subdirectory called openvpn '
                        'that contains the openvpn config files')
    g2.add_argument('--create-config', '-c', help=create_conf_help,
                    dest='create_conf_dir')

    # following args are used to support splitting clients among multiple VMs
    # each running vpn walker will use this to decide which portion of vpn
    # endpoints it should include
    parser.add_argument('--vm-num', dest='vm_num', type=int, default=1,
                        help="Specify the number of VMs running concurrently")
    parser.add_argument('--vm-index', dest='vm_index', type=int, default=1,
                        help='The index of current VM, must be >= 1 and '
                             '<= vm_num')
    return parser.parse_args()


def scan_vpns(directory, auth_file, crt_file, tls_auth, key_direction,
              exclude_list, shuffle_lists, vm_num, vm_index, reduce_vp):
    """
    For each VPN, check if there are experiments and scan with it if
    necessary

    Note: the expected directory structure is
    args.directory
    -----vpns (contains the OpenVPN config files
    -----configs (contains the Centinel config files)
    -----exps (contains the experiments directories)

    :param directory: root directory that contains vpn configs and
                      centinel client configs
    :param auth_file: a text file with username at first line and
                      password at second line
    :param crt_file: optional root certificate file
    :param tls_auth: additional key
    :param key_direction: must specify if tls_auth is used
    :param exclude_list: optional list of exluded countries
    :param shuffle_lists: shuffle vpn list if set true
    :param vm_num: number of VMs that are running currently
    :param vm_index: index of current VM
    :param reduce_vp: reduce number of vantage points
    :return:
    """

    logging.info("Starting to run the experiments for each VPN")
    logging.warn("Excluding vantage points from: %s" % exclude_list)

    # iterate over each VPN
    vpn_dir = return_abs_path(directory, "vpns")
    conf_dir = return_abs_path(directory, "configs")
    home_dir = return_abs_path(directory, "home")
    if auth_file is not None:
        auth_file = return_abs_path(directory, auth_file)
    if crt_file is not None:
        crt_file = return_abs_path(directory, crt_file)
    if tls_auth is not None:
        tls_auth = return_abs_path(directory, tls_auth)
    conf_list = sorted(os.listdir(conf_dir))

    # determine VPN provider
    vpn_provider = None
    if "hma" in directory:
        vpn_provider = "hma"
    elif "ipvanish" in directory:
        vpn_provider = "ipvanish"
    elif "purevpn" in directory:
        vpn_provider = "purevpn"
    elif "vpngate" in directory:
        vpn_provider = "vpngate"
    if vpn_provider:
        logging.info("Detected VPN provider is %s" % vpn_provider)
    else:
        logging.warning("Cannot determine VPN provider!")

    # reduce size of list if reduce_vp is true
    if reduce_vp:
        logging.info("Reducing list size. Original size: %d" % len(conf_list))
        country_asn_set = set()
        reduced_conf_set = set()
        for filename in conf_list:
            centinel_config = os.path.join(conf_dir, filename)
            config = centinel.config.Configuration()
            config.parse_config(centinel_config)
            vp_ip = os.path.splitext(filename)[0]

            try:
                meta = centinel.backend.get_meta(config.params, vp_ip)
                if 'country' in meta and 'as_number' in meta \
                        and meta['country'] and meta['as_number']:
                    country_asn = '_'.join([meta['country'], meta['as_number']])
                    if country_asn not in country_asn_set:
                        country_asn_set.add(country_asn)
                        reduced_conf_set.add(filename)
                else:
                    # run this endpoint if missing info
                    reduced_conf_set.add(filename)
            except:
                logging.warning("Failed to geolocate %s" % vp_ip)
                reduced_conf_set.add(filename)

        conf_list = list(reduced_conf_set)
        logging.info("List size reduced. New size: %d" % len(conf_list))

    # sort file list to ensure the same filename sequence in each VM
    conf_list = sorted(conf_list)

    # only select its own portion according to vm_num and vm_index
    chunk_size = len(conf_list) / vm_num
    last_chunk_additional = len(conf_list) % vm_num
    start_pointer = 0 + (vm_index - 1) * chunk_size
    end_pointer = start_pointer + chunk_size
    if vm_index == vm_num:
        end_pointer += last_chunk_additional
    conf_list = conf_list[start_pointer:end_pointer]

    if shuffle_lists:
        shuffle(conf_list)

    number = 1
    total = len(conf_list)

    external_ip = get_external_ip()
    if external_ip is None:
        logging.error("No network connection, exiting...")
        return
    logging.info("Current external IP: %s" % (external_ip))

    # getting namesevers that should be excluded
    local_nameservers = dns.resolver.Resolver().nameservers

    for filename in conf_list:
        # Check network connection first
        time.sleep(5)
        logging.info("Checking network connectivity...")
        current_ip = get_external_ip()
        if current_ip is None:
            logging.error("Network connection lost!")
            break
        elif current_ip != external_ip:
            logging.error("VPN still connected! IP: %s" % current_ip)
            if len(openvpn.OpenVPN.connected_instances) == 0:
                logging.error("No active OpenVPN instance found! Exiting...")
                break
            else:
                logging.warn("Trying to disconnect VPN")
                for instance in openvpn.OpenVPN.connected_instances:
                    instance.stop()
                    time.sleep(5)

                current_ip = get_external_ip()
                if current_ip is None or current_ip != external_ip:
                    logging.error("Stopping VPN failed! Exiting...")
                    break

            logging.info("Disconnecting VPN successfully")

        # start centinel for this endpoint
        logging.info("Moving onto (%d/%d) %s" % (number, total, filename))

        number += 1
        vpn_config = os.path.join(vpn_dir, filename)
        centinel_config = os.path.join(conf_dir, filename)

        # before starting the VPN, check if there are any experiments
        # to run
        config = centinel.config.Configuration()
        config.parse_config(centinel_config)

        # assuming that each VPN config file has a name like:
        # [ip-address].ovpn, we can extract IP address from filename
        # and use it to geolocate and fetch experiments before connecting
        # to VPN.
        vpn_address, extension = os.path.splitext(filename)
        country = None
        try:
            meta = centinel.backend.get_meta(config.params,
                                             vpn_address)
            if 'country' in meta:
                country = meta['country']
        except:
            logging.exception("%s: Failed to geolocate %s" % (filename, vpn_address))

        if country and exclude_list and country in exclude_list:
            logging.info("%s: Skipping this server (%s)" % (filename, country))
            continue

        # try setting the VPN info (IP and country) to get appropriate
        # experiemnts and input data.
        try:
            centinel.backend.set_vpn_info(config.params, vpn_address, country)
        except Exception as exp:
            logging.exception("%s: Failed to set VPN info: %s" % (filename, exp))

        logging.info("%s: Synchronizing." % filename)
        try:
            centinel.backend.sync(config.params)
        except Exception as exp:
            logging.exception("%s: Failed to sync: %s" % (filename, exp))

        if not experiments_available(config.params):
            logging.info("%s: No experiments available." % filename)
            try:
                centinel.backend.set_vpn_info(config.params, vpn_address, country)
            except Exception as exp:
                logging.exception("Failed to set VPN info: %s" % exp)
            continue

        # add exclude_nameservers to scheduler
        sched_path = os.path.join(home_dir, filename, "experiments", "scheduler.info")
        if os.path.exists(sched_path):
            with open(sched_path, 'r+') as f:
                sched_info = json.load(f)
                for task in sched_info:
                    if "python_exps" in sched_info[task] and "baseline" in sched_info[task]["python_exps"]:
                        if "params" in sched_info[task]["python_exps"]["baseline"]:
                            sched_info[task]["python_exps"]["baseline"]["params"]["exclude_nameservers"] = \
                                local_nameservers
                        else:
                            sched_info[task]["python_exps"]["baseline"]["params"] = \
                                {"exclude_nameservers": local_nameservers}

                # write back to same file
                f.seek(0)
                json.dump(sched_info, f, indent=2)
                f.truncate()

        logging.info("%s: Starting VPN." % filename)

        vpn = openvpn.OpenVPN(timeout=60, auth_file=auth_file, config_file=vpn_config,
                              crt_file=crt_file, tls_auth=tls_auth, key_direction=key_direction)

        vpn.start()
        if not vpn.started:
            logging.error("%s: Failed to start VPN!" % filename)
            vpn.stop()
            time.sleep(5)
            continue

        logging.info("%s: Running Centinel." % filename)
        try:
            client = centinel.client.Client(config.params, vpn_provider)
            centinel.conf = config.params
            # do not use client logging config
            # client.setup_logging()
            client.run()
        except Exception as exp:
            logging.exception("%s: Error running Centinel: %s" % (filename, exp))

        logging.info("%s: Stopping VPN." % filename)
        vpn.stop()
        time.sleep(5)

        logging.info("%s: Synchronizing." % filename)
        try:
            centinel.backend.sync(config.params)
        except Exception as exp:
            logging.exception("%s: Failed to sync: %s" % (filename, exp))

        # try setting the VPN info (IP and country) to the correct address
        # after sync is over.
        try:
            centinel.backend.set_vpn_info(config.params, vpn_address, country)
        except Exception as exp:
            logging.exception("Failed to set VPN info: %s" % exp)


def return_abs_path(directory, path):
    """
    Unfortunately, Python is not smart enough to return an absolute
    path with tilde expansion, so I writing functionality to do this

    :param directory:
    :param path:
    :return:
    """
    if directory is None or path is None:
        return
    directory = os.path.expanduser(directory)
    return os.path.abspath(os.path.join(directory, path))


def get_external_ip():
    # pool of URLs that returns public IP
    url_list = ["https://wtfismyip.com/text",
                "http://ip.42.pl/raw",
                "http://myexternalip.com/raw",
                "https://api.ipify.org/"]

    from urllib2 import urlopen, URLError
    # try four urls in case some are unreachable
    for url in url_list:
        try:
            my_ip = urlopen(url, timeout=5).read().rstrip()
            return my_ip
        except URLError:
            logging.warning("Failed to connect to %s" % url)
            continue
    # return None if all failed
    return None


def signal_handler(signal, frame):
    logging.warn("SIGINT or SIGTERM received")
    if len(openvpn.OpenVPN.connected_instances) > 0:
        logging.warn("Disconnecting VPN")
        for instance in openvpn.OpenVPN.connected_instances:
            instance.stop()
    sys.exit(0)


def create_config_files(directory):
    """
    For each VPN file in directory/vpns, create a new configuration
    file and all the associated directories

    Note: the expected directory structure is
    args.directory
    -----vpns (contains the OpenVPN config files
    -----configs (contains the Centinel config files)
    -----exps (contains the experiments directories)
    -----results (contains the results)

    :param directory:
    """
    logging.info("Starting to create config files from openvpn files")

    vpn_dir = return_abs_path(directory, "vpns")
    conf_dir = return_abs_path(directory, "configs")
    os.mkdir(conf_dir)
    home_dirs = return_abs_path(directory, "home")
    os.mkdir(home_dirs)
    for filename in os.listdir(vpn_dir):
        configuration = centinel.config.Configuration()
        # setup the directories
        home_dir = os.path.join(home_dirs, filename)
        os.mkdir(home_dir)
        configuration.params['user']['centinel_home'] = home_dir
        exp_dir = os.path.join(home_dir, "experiments")
        os.mkdir(exp_dir)
        configuration.params['dirs']['experiments_dir'] = exp_dir
        data_dir = os.path.join(home_dir, "data")
        os.mkdir(data_dir)
        configuration.params['dirs']['data_dir'] = data_dir
        res_dir = os.path.join(home_dir, "results")
        os.mkdir(res_dir)
        configuration.params['dirs']['results_dir'] = res_dir

        log_file = os.path.join(home_dir, "centinel.log")
        configuration.params['log']['log_file'] = log_file
        login_file = os.path.join(home_dir, "login")
        configuration.params['server']['login_file'] = login_file
        configuration.params['user']['is_vpn'] = True

        configuration.params['server']['verify'] = True
        configuration.params['experiments']['tcpdump_params'] = ["-i", "tun0"]

        conf_file = os.path.join(conf_dir, filename)
        configuration.write_out_config(conf_file)


def experiments_available(config):
    logging.info("Starting to check for experiments with %s",
                 config['server']['server_url'])
    try:
        client = centinel.client.Client(config)
        if client.has_experiments_to_run():
            return True
    except Exception, exp:
        logging.exception("Unable to check schedule: %s", str(exp))

    return False


def run():
    if os.path.isfile(PID_FILE):
        with open(PID_FILE) as f:
            pid = f.read()
            print "Centinel already running (PID = %s)" % pid
            print "Lock file address: %s" % PID_FILE
        sys.exit(1)

    try:
        f = open(PID_FILE, "w")
        f.write("%d" % os.getpid())
        f.close()
    except Exception as exp:
        sys.stderr.write('Error to writing the'
                         ' lock file: %s\n' % exp)
        sys.exit(1)

    try:
        _run()
    except SystemExit:
        pass
    except KeyboardInterrupt:
        print "Keyboard interrupt received, exiting..."
    except Exception as exp:
        sys.stderr.write("%s" % exp)

    try:
        os.remove(PID_FILE)
    except Exception as exp:
        sys.stderr.write("Failed to remove lock file %s: %s" % (PID_FILE, exp))


def _run():
    """Entry point for all uses of centinel"""
    args = parse_args()

    # register signal handler
    signal.signal(signal.SIGTERM, signal_handler)
    signal.signal(signal.SIGINT, signal_handler)

    # set up logging
    log_formatter = logging.Formatter("%(asctime)s %(filename)s(line %(lineno)d) "
                                      "%(levelname)s: %(message)s")
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.INFO)
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(log_formatter)
    root_logger.addHandler(console_handler)

    # add file handler if specified
    if args.log_file:
        file_handler = logging.FileHandler(args.log_file)
        file_handler.setFormatter(log_formatter)
        root_logger.addHandler(file_handler)

    # check vm_num and vm_index value
    if args.vm_num < 1:
        print "vm_num value cannot be negative!"
        return
    if args.vm_index < 1 or args.vm_index > args.vm_num:
        print "vm_index value cannot be negative or greater than vm_num!"
        return

    if args.create_conf_dir:
        if args.create_HMA:
            hma_dir = return_abs_path(args.create_conf_dir, 'vpns')
            hma.create_config_files(hma_dir)
        elif args.create_IPVANISH:
            ipvanish_dir = return_abs_path(args.create_conf_dir, 'vpns')
            ipvanish.create_config_files(ipvanish_dir)
        elif args.create_PUREVPN:
            purevpn_dir = return_abs_path(args.create_conf_dir, 'vpns')
            purevpn.create_config_files(purevpn_dir)
        elif args.create_VPNGATE:
            vpngate_dir = return_abs_path(args.create_conf_dir, 'vpns')
            vpngate.create_config_files(vpngate_dir)
        # create the config files for the openvpn config files
        create_config_files(args.create_conf_dir)
    else:
        # sanity check tls_auth and key_direction
        if (args.tls_auth is not None and args.key_direction is None) or \
                (args.tls_auth is None and args.key_direction is not None):
            logging.error("tls_auth and key_direction must be specified "
                          "together!")
            return

        scan_vpns(directory=args.directory, auth_file=args.auth_file,
                  crt_file=args.crt_file, tls_auth=args.tls_auth,
                  key_direction=args.key_direction, exclude_list=args.exclude_list,
                  shuffle_lists=args.shuffle_lists, vm_num=args.vm_num,
                  vm_index=args.vm_index, reduce_vp=args.reduce_vp)

if __name__ == "__main__":
    run()



# Dependencies: install M2Crypto with sudo apt-get install
# python-m2crypto

import logging

try:
    import M2Crypto
    m2crypto_imported = True
except ImportError:
    logging.warning("M2Crypto could not be imported. "
                    "TLS fingerprinting will be disabled.")
    m2crypto_imported = False

import ssl
import socket
import threading
import time

MAX_THREAD_START_RETRY = 10
THREAD_START_DELAY = 3


def get_fingerprint(host, port=443, external=None, log_prefix=''):
    tls_error = None
    fingerprint_error = None
    cert = None

    logging.debug("%sGetting TLS certificate "
                  "for %s:%d." % (log_prefix, host, port))
    # set socket timeout to 10 seconds so that we don't have
    # to wait forever if the server doesn't respond
    socket.setdefaulttimeout(10)

    try:
        cert = ssl.get_server_certificate((host, port),
                                          ssl_version=ssl.PROTOCOL_TLSv1)
    # if this fails, there's a possibility that SSLv3 handshake was
    # attempted and rejected by the server. Use TLSv1 instead.
    except ssl.SSLError:
        # exception could also happen here
        try:

            # this uses the highest version SSL or TLS that both 
            # endpoints support
            cert = ssl.get_server_certificate((host, port),
                                              ssl_version=ssl.PROTOCOL_SSLv23)
        except Exception as exp:
            tls_error = str(exp)
    except Exception as exp:
        tls_error = str(exp)

    # this comes out as unicode, but m2crypto breaks if it gets
    # something other than a string, so convert to ascii
    if type(cert) == unicode:
        cert = cert.encode('ascii', 'ignore')

    if tls_error is None and m2crypto_imported:
        try:
            x509 = M2Crypto.X509.load_cert_string(cert,
                                                  M2Crypto.X509.FORMAT_PEM)
            fingerprint = x509.get_fingerprint('sha1')
        except Exception as exp:
            fingerprint_error = str(exp)

    if not m2crypto_imported:
        fingerprint_error = "M2Crypto could not be imported."

    # the external result is used when threading to store
    # the results in the list container provided.
    row = "%s:%s" % (host, port)

    # handle return value based on exception types
    if tls_error is None and fingerprint_error is None:
        if external is not None and type(external) is dict:
            external[row] = {"cert": cert,
                             "fingerprint": fingerprint.lower()}
        return fingerprint.lower(), cert
    elif tls_error is None and fingerprint_error is not None:
        if external is not None and type(external) is dict:
            external[row] = {"cert": cert,
                             "fingerprint_error": fingerprint_error}
        return fingerprint_error, cert
    else:
        if external is not None and type(external) is dict:
            external[row] = {"tls_error": tls_error,
                             "fingerprint_error": fingerprint_error}
        return fingerprint_error, tls_error


def get_fingerprint_batch(input_list, results={}, default_port=443,
                          delay_time=0.5, max_threads=100):
    """
    This is a parallel version of the TLS fingerprint primitive.

    :param input_list: the input is a list of host:ports.
    :param default_port: default port to use when no port specified
    :param delay_time: delay before starting each thread
    :param max_threads: maximum number of concurrent threads
    :return:
    """
    threads = []
    thread_error = False
    thread_wait_timeout = 200
    ind = 1
    total_item_count = len(input_list)
    for row in input_list:
        if len(row.split(":")) == 2:
            host, port = row.split(":")
        elif len(row.split(":")) == 1:
            host = row
            port = default_port
        else:
            continue

        port = int(port)
        wait_time = 0
        while threading.active_count() > max_threads:
            time.sleep(1)
            wait_time += 1
            if wait_time > thread_wait_timeout:
                thread_error = True
                break

        if thread_error:
            results["error"] = "Threads took too long to finish."
            break

        # add just a little bit of delay before starting the thread
        # to avoid overwhelming the connection.
        time.sleep(delay_time)

        log_prefix = "%d/%d: " % (ind, total_item_count)
        thread = threading.Thread(target=get_fingerprint,
                                  args=(host, port,
                                        results, log_prefix))
        ind += 1
        thread.setDaemon(1)

        thread_open_success = False
        retries = 0
        while not thread_open_success and retries < MAX_THREAD_START_RETRY:
            try:
                thread.start()
                threads.append(thread)
                thread_open_success = True
            except:
                retries += 1
                time.sleep(THREAD_START_DELAY)
                logging.error("%sThread start failed for %s, retrying... (%d/%d)" % (log_prefix, host, retries, MAX_THREAD_START_RETRY))

        if retries == MAX_THREAD_START_RETRY:
            logging.error("%sCan't start a new thread for %s after %d retries." % (log_prefix, host, retries))

    for thread in threads:
        thread.join(thread_wait_timeout)

    return results



import base64
import logging
import threading
import time
import random
import BeautifulSoup
import re
from urlparse import urlparse

from http_helper import ICHTTPConnection
from centinel.utils import user_agent_pool

REDIRECT_LOOP_THRESHOLD = 5
MAX_THREAD_START_RETRY = 10
THREAD_START_DELAY = 3

def meta_redirect(content):
    """
    Returns redirecting URL if there is a HTML refresh meta tag,
    returns None otherwise

    :param content: HTML content
    """
    decoded = content.decode("utf-8", errors="replace")

    try:
        soup = BeautifulSoup.BeautifulSoup(decoded)
    except Exception as e:
        return None

    result = soup.find("meta", attrs={"http-equiv": re.compile("^refresh$", re.I)})
    if result:
        try:
            wait, text = result["content"].split(";")
            text = text.strip()
            if text.lower().startswith("url="):
                url = text[4:]
                return url
        except:
            # there are normal meta tag with refresh that are not
            # redirect and don't have a URL in it
            pass
    return None


def _get_http_request(netloc, path="/", headers=None, ssl=False):
    """
    Actually gets the http. Moved this to it's own private method since
    it is called several times for following redirects

    :param host:
    :param path:
    :param headers:
    :param ssl:
    :return:
    """
    if ssl:
        port = 443
    else:
        port = 80

    host = netloc

    if len(netloc.split(":")) == 2:
        host, port = netloc.split(":")

    request = {"host": host,
               "port": port,
               "path": path,
               "ssl": ssl,
               "method": "GET"}
    if headers:
        request["headers"] = headers

    response = {}

    try:
        conn = ICHTTPConnection(host=host, port=port, timeout=10)

        conn.request(path, headers, ssl, timeout=10)
        response["status"] = conn.status
        response["reason"] = conn.reason
        response["headers"] = conn.headers
        body = conn.body

        try:
            response["body"] = body.encode('utf-8')
        except UnicodeDecodeError:
            # if utf-8 fails to encode, just use base64
            response["body.b64"] = body.encode('base64')

    except Exception as err:
        response["failure"] = str(err)

    result = {"response": response,
              "request": request}

    return result


def get_request(netloc, path="/", headers=None, ssl=False,
                external=None, url=None, log_prefix=''):
    http_results = {}

    # Add User-Agent string if not present in headers
    if headers is None:
        headers = {"User-Agent": random.choice(user_agent_pool)}
    elif type(headers) is dict and "User-Agent" not in headers:
        headers["user-Agent"] = random.choice(user_agent_pool)

    first_response = _get_http_request(netloc, path, headers, ssl)
    if "failure" in first_response["response"]:  # If there was an error, just ignore redirects and return
        first_response_information = {"redirect_count": 0,
                                      "redirect_loop": False,
                                      "full_url": url,
                                      "response": first_response["response"],
                                      "request": first_response["request"]}
        http_results = first_response_information

        if external is not None and type(external) is dict:
            external[url] = http_results
        return http_results

    logging.debug("%sSending HTTP GET request for %s." % (log_prefix, url))

    response_headers_contains_location = False
    location_url = None
    # Checks HTTP Status code and location header to see if the webpage calls for a redirect
    for header, header_value in first_response["response"]["headers"].items():
        if header.lower() == "location":
            response_headers_contains_location = True
            location_url = header_value

    # check meta redirect
    meta_redirect_url = None
    is_meta_redirect = False
    if "body" in first_response["response"]:
        try:
            meta_redirect_url = meta_redirect(first_response["response"]["body"])
        except:
            logging.warning("%sError looking for redirects in: %s." % (log_prefix, url))
    elif "body.b64" in first_response["response"]:
        body_decoded = base64.b64decode(first_response["response"]["body.b64"])
        meta_redirect_url = meta_redirect(body_decoded)

    if meta_redirect_url is not None:
        is_meta_redirect = True

    is_redirecting = response_headers_contains_location or is_meta_redirect


    previous_url = ""
    previous_netloc = netloc
    if is_redirecting:
        http_results["redirects"] = {}
        first_response_information = {"full_url": url,
                                      "response": first_response["response"],
                                      "request": first_response["request"]}
        http_results["redirects"][0] = first_response_information
        redirect_http_result = None
        redirect_number = 1
        redirect_failed = False
        while redirect_http_result is None or is_redirecting and\
                not redirect_failed and\
                redirect_number <= REDIRECT_LOOP_THRESHOLD:  # While there are more redirects...
            # Usually, redirects that redirect more than 5 times are infinite loops
            if response_headers_contains_location:
                redirect_url = location_url
            elif is_meta_redirect:
                redirect_url = meta_redirect_url

            # prevent looping on the same URL
            if previous_url == redirect_url:
                break
            previous_url = redirect_url

            use_ssl = redirect_url.startswith("https://")  # If redirect url starts with https, use ssl

            # Scheme, query, and fragment aren't used. Urlparse is used here to split the url into the host and path
            # Useful for httplib since it requires this
            parsed_url = urlparse(redirect_url)

            netloc = parsed_url.netloc
            # if host is not specified, use the last one
            if netloc is None or netloc == "":
                netloc = previous_netloc

            previous_netloc = netloc

            redirect_http_result = _get_http_request(netloc, parsed_url.path, ssl=use_ssl)

            # If there is an error in the redirects, break the loop and stop there
            if "failure" in redirect_http_result["response"]:
                redirect_failed = True

            response_headers_contains_location = False
            location_url = None
            is_meta_redirect = False
            if not redirect_failed:
                for header, header_value in redirect_http_result["response"]["headers"].items():
                    if header.lower() == "location":
                        response_headers_contains_location = True
                        location_url = header_value

                # check meta redirect
                meta_redirect_url = None
                if "body" in redirect_http_result["response"]:
                    meta_redirect_url = meta_redirect(redirect_http_result["response"]["body"])
                elif "body.b64" in redirect_http_result["response"]:
                    body_decoded = base64.b64decode(first_response["response"]["body.b64"])
                    meta_redirect_url = meta_redirect(body_decoded)

                if meta_redirect_url is not None:
                    is_meta_redirect = True

            is_redirecting = response_headers_contains_location or is_meta_redirect

            # If this is the final response, put this in the first request and response json
            if not is_redirecting or redirect_number == REDIRECT_LOOP_THRESHOLD:
                http_results["redirect_loop"] = (is_redirecting and redirect_number ==  REDIRECT_LOOP_THRESHOLD)
                http_results["redirect_count"] = redirect_number
                http_results["full_url"] = redirect_url

            redirect_information = {"full_url": redirect_url,
                                    "response": redirect_http_result["response"],
                                    "request": redirect_http_result["request"]}
            http_results["redirects"][redirect_number] = redirect_information

            redirect_number += 1

    else:
        first_response_information = {"redirect_count": 0,
                                      "redirect_loop": False,
                                      "full_url": url,
                                      "response": first_response["response"],
                                      "request": first_response["request"]}
        http_results = first_response_information

    # the external result is used when threading to store
    # the results in the list container provided.
    if external is not None and type(external) is dict:
        external[url] = http_results
    return http_results


def get_requests_batch(input_list, results={}, delay_time=0.5, max_threads=100):
    """
    This is a parallel version of the HTTP GET primitive.

    :param input_list: the input is a list of either dictionaries containing
                       query information, or just domain names (and NOT URLs).
    :param delay_time: delay before starting each thread
    :param max_threads: maximum number of concurrent threads
    :return: results in dict format

    Note: the input list can look like this:
    [
        { "host": "www.google.com",   "path": "/", "headers": {},
          "ssl": False, "url": "http://www.google.com/" },
        "www.twitter.com",
        "www.youtube.com",
        { "host": "www.facebook.com", "path": "/", "headers": {},
          "ssl": True, "url": "http://www.facebook.com" },
        ...
    ]
    """
    threads = []
    thread_error = False
    thread_wait_timeout = 200
    ind = 1
    total_item_count = len(input_list)
    # randomly select one user agent for one input list
    user_agent = random.choice(user_agent_pool)
    for row in input_list:
        headers = {}
        path = "/"
        ssl = False
        theme = "http"
        if type(row) is dict:
            if "host" not in row:
                continue
            host = row["host"]

            if "path" in row:
                path = row["path"]

            if "headers" in row:
                if type(row["headers"]) is dict:
                    headers = row["headers"]

            if "ssl" in row:
                ssl = row["ssl"]
                theme = "https"

            if "url" in row:
                url = row["url"]
            else:
                url = "%s://%s%s" % (theme, host, path)
        else:
            host = row
            url = "%s://%s%s" % (theme, host, path)

        wait_time = 0
        while threading.active_count() > max_threads:
            time.sleep(1)
            wait_time += 1
            if wait_time > thread_wait_timeout:
                thread_error = True
                break

        if thread_error:
            results["error"] = "Threads took too long to finish."
            break

        if "User-Agent" not in headers:
            headers["User-Agent"] = user_agent

        # add just a little bit of delay before starting the thread
        # to avoid overwhelming the connection.
        time.sleep(delay_time)
        log_prefix = "%d/%d: " % (ind, total_item_count)
        thread = threading.Thread(target=get_request,
                                  args=(host, path, headers, ssl,
                                        results, url, log_prefix))
        ind += 1
        thread.setDaemon(1)

        thread_open_success = False
        retries = 0
        while not thread_open_success and retries < MAX_THREAD_START_RETRY:
            try:
                thread.start()
                threads.append(thread)
                thread_open_success = True
            except:
                retries += 1
                time.sleep(THREAD_START_DELAY)
                logging.error("%sThread start failed for %s, retrying... (%d/%d)" % (log_prefix, url, retries, MAX_THREAD_START_RETRY))

        if retries == MAX_THREAD_START_RETRY:
            logging.error("%sCan't start a new thread for %s after %d retries." % (log_prefix, url, retries))

    for thread in threads:
        thread.join(thread_wait_timeout)

    return results



from base64 import b64encode
import dns.rdatatype
import dns.message
import dns.resolver
import logging
import select
import socket
import threading
import time

MAX_THREAD_START_RETRY = 10
THREAD_START_DELAY = 3

def get_ips(host, nameserver=None, record="A"):
    nameservers = []
    if nameserver is not None:
        nameservers = [nameserver]
    return lookup_domain(host, nameservers=nameservers, rtype=record)


def lookup_domain(domain, nameservers=[], rtype="A",
                  exclude_nameservers=[], timeout=2):
    """Wrapper for DNSQuery method"""
    dns_exp = DNSQuery(domains=[domain], nameservers=nameservers, rtype=rtype,
                       exclude_nameservers=exclude_nameservers, timeout=timeout)
    return dns_exp.lookup_domain(domain)


def lookup_domains(domains, results={}, nameservers=[], exclude_nameservers=[],
                   rtype="A", timeout=2):
    dns_exp = DNSQuery(domains=domains, results=results, nameservers=nameservers, 
                       rtype=rtype, exclude_nameservers=exclude_nameservers, 
                       timeout=timeout)
    return dns_exp.lookup_domains()


def send_chaos_queries():
    dns_exp = DNSQuery()
    return dns_exp.send_chaos_queries()


class DNSQuery:
    """Class to store state for all of the DNS queries"""

    def __init__(self, domains=[], results={}, nameservers=[], exclude_nameservers=[],
                 rtype="A", timeout=10, max_threads=100):
        """Constructor for the DNS query class

        Params:
        nameserver- the nameserver to use, defaults to the local resolver
        rtype- the record type to lookup (as text), by default A
        timeout- how long to wait for a response, by default 10 seconds

        """
        self.domains = domains
        self.results = results
        self.rtype = rtype
        self.timeout = timeout
        self.max_threads = max_threads
        if len(nameservers) == 0:
            nameservers = dns.resolver.Resolver().nameservers
        # remove excluded nameservers
        if len(exclude_nameservers) > 0:
            for nameserver in exclude_nameservers:
                if nameserver in nameservers:
                    nameservers.remove(nameserver)
        # include google nameserver
        if "8.8.8.8" not in nameservers:
            nameservers.append("8.8.8.8")
        self.nameservers = nameservers
        self.threads = []
        # start point of port number to be used
        self.port = 30000
        # create thread lock for port number index
        self.port_lock = threading.Lock()

    def send_chaos_queries(self):
        """Send chaos queries to identify the DNS server and its manufacturer

        Note: we send 2 queries for BIND stuff per RFC 4892 and 1
        query per RFC 6304

        Note: we are not waiting on a second response because we
        shouldn't be getting injected packets here

        """
        names = ["HOSTNAME.BIND", "VERSION.BIND", "ID.SERVER"]
        self.results = {'exp-name': "chaos-queries"}
        for name in names:
            self.results[name] = {}
            for nameserver in self.nameservers:
                sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                sock.settimeout(self.timeout)
                query = dns.message.make_query(name,
                                               dns.rdatatype.from_text("TXT"),
                                               dns.rdataclass.from_text("CH"))
                sock.sendto(query.to_wire(), (nameserver, 53))
                reads, _, _ = select.select([sock], [], [], self.timeout)
                if len(reads) == 0:
                    self.results[name][nameserver] = None
                else:
                    response = reads[0].recvfrom(4096)[0]
                    self.results[name][nameserver] = b64encode(response)
        return self.results

    def lookup_domains(self):
        """More complex DNS primitive that looks up domains concurrently

        Note: if you want to lookup multiple domains, you should use
        this function
        """
        thread_error = False
        thread_wait_timeout = 200
        ind = 1
        total_item_count = len(self.domains)
        for domain in self.domains:
            for nameserver in self.nameservers:
                wait_time = 0
                while threading.active_count() > self.max_threads:
                    time.sleep(1)
                    wait_time += 1
                    if wait_time > thread_wait_timeout:
                        thread_error = True
                        break

                if thread_error:
                    self.results["error"] = "Threads took too long to finish."
                    break
                log_prefix = "%d/%d: " % (ind, total_item_count)
                thread = threading.Thread(target=self.lookup_domain,
                                          args=(domain, nameserver,
                                                log_prefix))
                thread.setDaemon(1)

                thread_open_success = False
                retries = 0
                while not thread_open_success and retries < MAX_THREAD_START_RETRY:
                    try:
                        thread.start()
                        self.threads.append(thread)
                        thread_open_success = True
                    except:
                        retries += 1
                        time.sleep(THREAD_START_DELAY)
                        logging.error("%sThread start failed for %s, retrying... (%d/%d)" % (log_prefix, domain, retries, MAX_THREAD_START_RETRY))

                if retries == MAX_THREAD_START_RETRY:
                    logging.error("%sCan't start a new thread for %s after %d retries." % (log_prefix, domain, retries))

            if thread_error:
                break
            ind += 1

        for thread in self.threads:
            thread.join(self.timeout * 3)
        return self.results

    def lookup_domain(self, domain, nameserver=None, log_prefix=''):
        """Most basic DNS primitive that looks up a domain, waits for a
        second response, then returns all of the results

        :param domain: the domain to lookup
        :param nameserver: the nameserver to use
        :param log_prefix:
        :return:

        Note: if you want to lookup multiple domains you *should not* use
        this function, you should use lookup_domains because this does
        blocking IO to wait for the second response

        """
        if domain not in self.results:
            self.results[domain] = []
        # get the resolver to use
        if nameserver is None:
            logging.debug("Nameserver not specified, using %s" % self.nameservers[0])
            nameserver = self.nameservers[0]
        results = {'domain': domain, 'nameserver': nameserver}
        # construct the socket to use
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.settimeout(self.timeout)
        # set port number and increment index:
        interrupt = False
        with self.port_lock:
            counter = 1
            while True:
                if counter > 100:
                    logging.warning("Stop trying to get an available port")
                    interrupt = True
                    break
                try:
                    sock.bind(('', self.port))
                    break
                except socket.error:
                    logging.debug("Port {} already in use, try next one".format(self.port))
                    self.port += 1
                    counter += 1
            self.port += 1

        if interrupt:
            sock.close()
            results['error'] = 'Failed to run DNS test'
            self.results[domain].append(results)
            return results

        logging.debug("%sQuerying DNS enteries for "
                      "%s (nameserver: %s)." % (log_prefix, domain, nameserver))

        # construct and send the request
        request = dns.message.make_query(domain,
                                         dns.rdatatype.from_text(self.rtype))
        results['request'] = b64encode(request.to_wire())
        sock.sendto(request.to_wire(), (nameserver, 53))

        # read the first response from the socket
        try:
            response = sock.recvfrom(4096)[0]
            results['response1'] = b64encode(response)
            resp = dns.message.from_wire(response)
            results['response1-ips'] = parse_out_ips(resp)

            # first domain name in response should be the same with query
            # domain name
            for entry in resp.answer:
                if domain.lower() != entry.name.to_text().lower()[:-1]:
                    logging.debug("%sWrong domain name %s for %s!"
                                  % (log_prefix, entry.name.to_text().lower()[:-1], domain))
                    results['response1-domain'] = entry.name.to_text().lower()[:-1]
                break
        except socket.timeout:
            # if we didn't get anything, then set the results to nothing
            logging.debug("%sQuerying DNS enteries for "
                          "%s (nameserver: %s) timed out!" % (log_prefix, domain, nameserver))
            sock.close()
            results['response1'] = None
            self.results[domain].append(results)
            return results

        # if we have made it this far, then wait for the next response
        try:
            response2 = sock.recvfrom(4096)[0]
            results['response2'] = b64encode(response2)
            resp2 = dns.message.from_wire(response2)
            results['response2-ips'] = parse_out_ips(resp2)

            # first domain name in response should be the same with query
            # domain name
            for entry in resp2.answer:
                if domain.lower != entry.name.to_text().lower()[:-1]:
                    logging.debug("%sWrong domain name %s for %s!"
                                  % (log_prefix, entry.name.to_text().lower()[:-1], domain))
                    results['response2-domain'] = entry.name.to_text().lower()[:-1]
                break
        except socket.timeout:
            # no second response
            results['response2'] = None

        sock.close()
        self.results[domain].append(results)
        return results


def parse_out_ips(message):
    """Given a message, parse out the ips in the answer"""

    ips = []
    for entry in message.answer:
        for rdata in entry.items:
            ips.append(rdata.to_text())
    return ips



import logging
import os
import shutil
import socket
import sys
import urllib
import zipfile


def unzip(source_filename, dest_dir):
    with zipfile.ZipFile(source_filename) as zf:
        zf.extractall(dest_dir)


def create_config_files(directory):
    """
    Initialize directory ready for vpn walker
    :param directory: the path where you want this to happen
    :return:
    """
    # Some constant strings
    config_zip_url = "https://s3-us-west-1.amazonaws.com/heartbleed/linux/linux-files.zip"

    if not os.path.exists(directory):
        os.makedirs(directory)

    logging.info("Starting to download PureVPN config file zip")
    url_opener = urllib.URLopener()
    zip_path = os.path.join(directory, '../linux_files.zip')
    url_opener.retrieve(config_zip_url, zip_path)
    logging.info("Extracting zip file")
    unzip(zip_path, os.path.join(directory, '../'))

    # remove zip file
    os.remove(zip_path)
    # copy ca and key to root path
    shutil.copyfile(os.path.join(directory, '../Linux OpenVPN Updated files', 'ca.crt'),
                    os.path.join(directory, '../ca.crt'))
    shutil.copyfile(os.path.join(directory, '../Linux OpenVPN Updated files', 'Wdc.key'),
                    os.path.join(directory, '../Wdc.key'))
    # move all config files to /vpns
    orig_path = os.path.join(directory, '../Linux OpenVPN Updated files/TCP')

    server_country = {}
    for filename in os.listdir(orig_path):
        if filename.endswith('.ovpn'):
            country = filename.split('-')[0]
            if '(V)' in country:
                country = country[:country.find('(V)')]

            file_path = os.path.join(orig_path, filename)
            lines = [line.rstrip('\n') for line in open(file_path)]

            # get ip address for this vpn
            ip = ""
            for line in lines:
                if line.startswith('remote'):
                    hostname = line.split(' ')[1]
                    ip = socket.gethostbyname(hostname)
                    break

            if len(ip) > 0:
                new_path = os.path.join(directory, ip + '.ovpn')
                shutil.copyfile(file_path, new_path)
                server_country[ip] = country

    # remove extracted folder
    shutil.rmtree(os.path.join(directory, '../Linux OpenVPN Updated files'))

    # add dns update options to each file
    logging.info("Appending DNS update options")
    for filename in os.listdir(directory):
        file_path = os.path.join(directory, filename)
        with open(file_path, 'a') as f:
            f.write("\n")
            f.write("up /etc/openvpn/update-resolv-conf\n")
            f.write("down /etc/openvpn/update-resolv-conf\n")

    print os.path.join(directory, 'servers.txt'), len(server_country)
    with open(os.path.join(directory, 'servers.txt'), 'w') as f:
        for ip in server_country:
            f.write('|'.join([ip, server_country[ip]]) + '\n')


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print "Usage {0} <directory to create VPNs in>".format(sys.argv[0])
        sys.exit(1)
    create_config_files(sys.argv[1])



#!/usr/bin/python
# Ben Jones bjones99@gatech.edu
# Georgia Tech Fall 2014
#
# command.py: code to manage external programs with subprocess

import os
import signal
import subprocess
import threading
import time


class Command():
    """Class to handle the interface between Python scripts and executables"""

    def __init__(self, command, output_callback, timeout=10):
        """Constructor for the command class that sets up generic
        timed out execution

        Params:
        command- the command to run in the form of a list of strings
            (per subprocess input

        output_callback(self, line, kill_switch)- the callback to
            process what the program prints to stdout and stderr where
            line is the latest line (all output is also logged to
            self.notifications) This function should also set the state
            variables self.started, self.error, and self.stopped
            correctly. The function may assume that these are
            initialized to False, but must appropriately change the
            variables depending on input. On an error, the function
            must also set self.error to True and self.error_msg to the
            error message

        """
        self.command = command
        self.output_callback = output_callback
        self.timeout = timeout
        self.started = False
        self.stopped = False
        self.exception = None
        self.error = False
        self.notifications = ""

        self.thread = threading.Thread(target=self._invoke_cmd)
        self.thread.setDaemon(1)

    def start(self, timeout=None):
        """Start running the command"""

        self.thread.start()
        start_time = time.time()
        if not timeout:
            timeout = self.timeout
        # every second, check the condition of the thread and return
        # control to the user if appropriate
        while start_time + timeout > time.time():
            self.thread.join(1)
            if self.started:
                return True
            if self.error:
                return False
        return False

    def stop(self, timeout=None):
        """Stop the given command"""

        if not timeout:
            timeout = self.timeout
        self.kill_switch()
        # Send the signal to all the process groups
        self.process.kill()
        self.thread.join(timeout)
        try:
            os.killpg(os.getpgid(self.process.pid), signal.SIGTERM)
        except:
            pass
        if self.stopped:
            return True
        else:
            return False

    def _invoke_cmd(self):
        # if the command execution throws an exception,
        # it should be caught and stored in a variable.
        try:
            self.process = subprocess.Popen(self.command,
                                            stdin=subprocess.PIPE,
                                            stdout=subprocess.PIPE,
                                            stderr=subprocess.STDOUT,
                                            preexec_fn=os.setsid)
        except Exception as exp:
            self.exception = exp
            self.started = False
            self.error = False
            return

        self.kill_switch = self.process.terminate
        self.starting = True
        while True:
            line = self.process.stdout.readline().strip()
            if not line:
                break
            self.output_callback(self, line, self.process.terminate)
            self.notifications += line + "\n"



class ExperimentList(type):
    experiments = {}

    def __init__(cls, name, bases, attrs):
        if name != "Experiment":
            ExperimentList.experiments[cls.name] = cls


class Experiment:
    __metaclass__ = ExperimentList

    # a list of input files that can be
    # used in order to make use of more than
    # one input file. If none specified, the
    # [experiment name].txt will be used instead.
    input_files = None

    # this should be set to True if the experiment
    # does its own tcpdump recording.
    overrides_tcpdump = False

    # if the experiment produces files that are not
    # to be included in the json file, it should
    # keep them in this dictionary.
    # { "file1_name.extention" : "[file1_contents]",
    #   "file2_name.extention" : "[file2_contents]",
    #   ... }
    # these files will be compressed when being stored
    external_results = None

    # an experiment can have external parameters
    # that are usually set by the scheduler.
    params = {}

    def run(self):
        raise NotImplementedError



import StringIO
import csv
import logging
import os
import requests
import sys
from base64 import b64decode


def create_config_files(directory):
    """
    Initialize directory ready for vpn walker
    :param directory: the path where you want this to happen
    :return:
    """
    # Some constant strings
    vpn_gate_url = "http://www.vpngate.net/api/iphone/"

    if not os.path.exists(directory):
        os.makedirs(directory)

    # get csv into memory
    csv_str = ""
    logging.info("Downloading info from VPN Gate API...")
    r = requests.get(vpn_gate_url)
    for line in r.text.split('\n'):
        csv_str += line.encode('utf-8')
        csv_str += "\n"

    # convert csv string to string IO
    f = StringIO.StringIO(csv_str)

    # generate vpn dict
    vpn_dict = {}
    reader = csv.reader(f)
    reader.next()
    reader.next()
    for row in reader:
        if len(row) == 15:
            alpha2 = row[6]
            vpn_dict[alpha2] = vpn_dict.get(alpha2, [])
            vpn_dict[alpha2].append({
                "vpn_name": row[0],
                "ip": row[1],
                "country_name": row[5],
                "alpha2": alpha2,
                "openvpn_config": b64decode(row[-1])
            })
    f.close()

    server_country = {}

    # write config files
    for country in vpn_dict:
        for data in vpn_dict[country]:
            config_filename = "{}.ovpn".format(data['ip'])
            file_path = os.path.join(directory, config_filename)
            with open(file_path, 'w') as f:
                f.write(data['openvpn_config'])
                f.write("up /etc/openvpn/update-resolv-conf\n")
                f.write("down /etc/openvpn/update-resolv-conf\n")

            server_country[data['ip']] = country

    with open(os.path.join(directory, 'servers.txt'), 'w') as f:
        for ip in server_country:
            f.write('|'.join([ip, server_country[ip]]) + '\n')


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print "Usage {0} <directory to create VPNs in>".format(sys.argv[0])
        sys.exit(1)
    create_config_files(sys.argv[1])



#!/usr/bin/env python
import argparse
import logging
import getpass
import os
import os.path
import sys

import centinel
import centinel.config
import centinel.daemonize

PID_FILE = "/tmp/centinel.lock"

# Constants
DEFAULT_CONFIG_FILE = os.path.expanduser('~' + getpass.getuser() +
                                         "/.centinel/config.ini")


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--config', '-c', help='Configuration file',
                        dest='config')
    no_verify_help = ("Disable certificate verification (NOT RECOMMENDED! "
                      "Use only when debugging.)")
    parser.add_argument('--no-verify', '-nv', help=no_verify_help,
                        action='store_true')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--version', '-v', action='version',
                       version="Centinel %s" % centinel.__version__,
                       help='Print the installed version number')
    group.add_argument('--sync', help='Sync data with server',
                       action='store_true')
    consent_help = ("Give informed consent so that you can download "
                    "experiments from the researchers at ICLab and upload "
                    "results for analysis")
    group.add_argument('--informed-consent', help=consent_help,
                       dest='consent', default=False, action='store_true')
    daemon_help = ('Create cron jobs to run Centinel in the background and '
                   'autoupdate the package. You need to run this command as '
                   'root in order to setup this functionality. '
                   'By default, this will use /usr/local/bin/centinel '
                   'for the binary location and will create an autoupdate '
                   'script.')
    parser.add_argument('--daemonize', help=daemon_help, action='store_true',
                        dest='daemonize')
    user_help = ('Using this option with --daemonize will make the '
                 'cron job created to run Centinel as the specified user '
                 'instead of root. You will still need to run this '
                 'command as root to setup this functionality. '
                 'By default, root is used to run both daemonized scripts.')
    parser.add_argument('--user', help=user_help, default="root")
    verbose_help = ('Verbose logging')
    parser.add_argument('--verbose', '-V', help=verbose_help,
                        action='store_true', dest='verbose')
    binary_help = ('Name or location of the binary to use in the cron job '
                   'for Centinel')
    parser.add_argument('--binary', help=binary_help,
                        default='/usr/local/bin/centinel')
    update_help = ('Create an autoupdate script for the installed package. '
                   'Note that you must have installed from a pip package for '
                   'this to work correctly and you must also set the '
                   'daemonize option')
    parser.add_argument('--auto-update', action='store_false',
                        help=update_help, default="centinel")
    group.add_argument('--update-config', help='Update configuration file',
                       action='store_true')
    meta_help = ('Add custom meta tags to results. '
                 'Format is: key1:value1,key2:value2,... without spaces')
    parser.add_argument('--meta', '-m', help=meta_help, dest='custom_meta')

    args = parser.parse_args()
    if (not args.daemonize and 
        (args.auto_update != 'centinel' or
         args.binary != '/usr/local/bin/centinel' or
         args.user != "root")):
        parser.error("--auto-update, --user, and --binary must be used with "
                     "--daemonize")
    return args


def run():
    if os.path.isfile(PID_FILE):
        with open(PID_FILE) as f:
            pid = f.read()
            print "Centinel already running (PID = %s)" % pid
            print "Lock file address: %s" % PID_FILE
        sys.exit(1)

    try:
        f = open(PID_FILE, "w")
        f.write("%d" % os.getpid())
        f.close()
    except Exception as exp:
        sys.stderr.write('Error to writing the'
                         ' lock file: %s\n' % exp)
        sys.exit(1)

    try:
        _run()
    except SystemExit:
        pass
    except KeyboardInterrupt:
        print "Keyboard interrupt received, exiting..."
    except Exception as exp:
        sys.stderr.write("%s" % exp)

    try:
        os.remove(PID_FILE)
    except Exception as exp:
        sys.stderr.write("Failed to remove lock file %s: %s\n" % (PID_FILE, exp))


def _run():
    """Entry point for package and cli uses"""

    args = parse_args()

    # parse custom parameters
    custom_meta = None
    if args.custom_meta:
        print "Adding custom parameters:"
        custom_meta = {}
        try:
            for item in args.custom_meta.split(','):
                key, value = item.split(':')
                custom_meta[key] = value
                print 'key: %s, value: %s' % (key, value)
        except Exception as e:
            sys.stderr.write("ERROR: Can not parse custom meta tags! %s\n" % (str(e)))

    # we need to store some persistent info, so check if a config file
    # exists (default location is ~/.centinel/config.ini). If the file
    # does not exist, then create a new one at run time
    configuration = centinel.config.Configuration()
    if args.config:
        configuration.parse_config(args.config)
    else:
        # if the file does not exist, then the default config file
        # will be used
        new_configuration = None
        if os.path.exists(DEFAULT_CONFIG_FILE):
            configuration.parse_config(DEFAULT_CONFIG_FILE)
        else:
            print 'Configuration file does not exist. Creating a new one.'
            new_configuration = centinel.config.Configuration()

        if not ('version' in configuration.params and
                configuration.params['version']['version'] == centinel.__version__):
            if not args.update_config:
                print ('WARNING: configuration file is from '
                       'a different version (%s) of '
                       'Centinel. Run with --update-config to update '
                       'it.' % (configuration.params['version']['version']))
            else:
                new_configuration = centinel.config.Configuration()
                backup_path = DEFAULT_CONFIG_FILE + ".old"
                new_configuration.update(configuration, backup_path)

        if new_configuration is not None:
            configuration = new_configuration
            configuration.write_out_config(DEFAULT_CONFIG_FILE)
            print 'New configuration written to %s' % (DEFAULT_CONFIG_FILE)
            if args.update_config:
                sys.exit(0)

    if args.verbose:
        if 'log' not in configuration.params:
            configuration.params['log'] = dict()
        configuration.params['log']['log_level'] = logging.DEBUG

    # add custom meta values from CLI
    if custom_meta is not None:
        if 'custom_meta' in configuration.params:
            configuration.params['custom_meta'].update(custom_meta)
        else:
            configuration.params['custom_meta'] = custom_meta

    centinel.conf = configuration.params
    client = centinel.client.Client(configuration.params)
    client.setup_logging()
    # disable cert verification if the flag is set
    if args.no_verify:
        configuration.params['server']['verify'] = False

    user = centinel.backend.User(configuration.params)
    # Note: because we have mutually exclusive arguments, we don't
    # have to worry about multiple arguments being called
    if args.sync:
        centinel.backend.sync(configuration.params)
    elif args.consent:
        user.informed_consent()
    elif args.daemonize:
        # if we don't have a valid binary location, then exit
        if not os.path.exists(args.binary):
            print "Error: no binary found to daemonize"
            exit(1)
        centinel.daemonize.daemonize(args.auto_update, args.binary,
            args.user)
    else:
        client.run()

if __name__ == "__main__":
    run()



#!/usr/bin/python
# openvpn.py: library to handle starting and stopping openvpn instances

import logging
import os
import signal
import subprocess
import threading
import time


class OpenVPN:
    connected_instances = []

    def __init__(self, config_file=None, auth_file=None, crt_file=None,
                 tls_auth=None, key_direction=None, timeout=60):
        self.started = False
        self.stopped = False
        self.error = False
        self.notifications = ""
        self.auth_file = auth_file
        self.crt_file = crt_file
        self.tls_auth = tls_auth
        self.key_dir = key_direction
        self.config_file = config_file
        self.thread = threading.Thread(target=self._invoke_openvpn)
        self.thread.setDaemon(1)
        self.timeout = timeout

    def _invoke_openvpn(self):
        cmd = ['sudo', 'openvpn', '--script-security', '2']
        # --config must be the first parameter, since otherwise
        # other specified options might not be able to overwrite
        # the wrong, relative-path options in config file
        if self.config_file is not None:
            cmd.extend(['--config', self.config_file])
        if self.crt_file is not None:
            cmd.extend(['--ca', self.crt_file])
        if self.tls_auth is not None and self.key_dir is not None:
            cmd.extend(['--tls-auth', self.tls_auth, self.key_dir])
        if self.auth_file is not None:
            cmd.extend(['--auth-user-pass', self.auth_file])

        self.process = subprocess.Popen(cmd,
                                        stdin=subprocess.PIPE,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.STDOUT,
                                        preexec_fn=os.setsid)
        self.kill_switch = self.process.terminate
        self.starting = True
        while True:
            line = self.process.stdout.readline().strip()
            if not line:
                break
            self.output_callback(line, self.process.terminate)

    def output_callback(self, line, kill_switch):
        """Set status of openvpn according to what we process"""
        self.notifications += line + "\n"

        if "Initialization Sequence Completed" in line:
            self.started = True
        if "ERROR:" in line or "Cannot resolve host address:" in line:
            self.error = True
        if "process exiting" in line:
            self.stopped = True

    def start(self, timeout=None):
        """
        Start OpenVPN and block until the connection is opened or there is
        an error
        :param timeout: time in seconds to wait for process to start
        :return:
        """
        if not timeout:
            timeout = self.timeout
        self.thread.start()
        start_time = time.time()
        while start_time + timeout > time.time():
            self.thread.join(1)
            if self.error or self.started:
                break
        if self.started:
            logging.info("OpenVPN connected")
            # append instance to connected list
            OpenVPN.connected_instances.append(self)
        else:
            logging.warn("OpenVPN not started")
            for line in self.notifications.split('\n'):
                logging.warn("OpenVPN output:\t\t%s" % line)

    def stop(self, timeout=None):
        """
        Stop OpenVPN process group
        :param timeout: time in seconds to wait for process to stop
        :return:
        """
        if not timeout:
            timeout = self.timeout
        os.killpg(os.getpgid(self.process.pid), signal.SIGTERM)
        self.thread.join(timeout)
        if self.stopped:
            logging.info("OpenVPN stopped")
            if self in OpenVPN.connected_instances:
                OpenVPN.connected_instances.remove(self)
        else:
            logging.error("Cannot stop OpenVPN!")
            for line in self.notifications.split('\n'):
                logging.warn("OpenVPN output:\t\t%s" % line)



import shutil
import csv
import json
import logging
import os
import sys
import time

from pyvirtualdisplay import Display
from selenium import webdriver
from selenium.webdriver.firefox.firefox_binary import FirefoxBinary

import foctor_core.foctor_core as fc


class HeadlessBrowser:
    def __init__(self):
        self.cur_path = os.path.dirname(os.path.abspath(__file__))
        self.display = Display(visible=False)
        self.binary = None
        self.profile = None
        self.driver = None
        self.parsed = 0

    @fc.timing
    def setup_profile(self, firebug=True, netexport=True):
        """
        Setup the profile for firefox
        :param firebug: whether add firebug extension
        :param netexport: whether add netexport extension
        :return: a firefox profile object
        """
        profile = webdriver.FirefoxProfile()
        profile.set_preference("app.update.enabled", False)
        if firebug:
            profile.add_extension(os.path.join(self.cur_path, 'extensions/firebug-2.0.8.xpi'))
            profile.set_preference("extensions.firebug.currentVersion", "2.0.8")
            profile.set_preference("extensions.firebug.allPagesActivation", "on")
            profile.set_preference("extensions.firebug.defaultPanelName", "net")
            profile.set_preference("extensions.firebug.net.enableSites", True)
            profile.set_preference("extensions.firebug.delayLoad", False)
            profile.set_preference("extensions.firebug.onByDefault", True)
            profile.set_preference("extensions.firebug.showFirstRunPage", False)
            profile.set_preference("extensions.firebug.net.defaultPersist", True)  # persist all redirection responses
        if netexport:
            har_path = os.path.join(self.cur_path, "har")
            if not os.path.exists(har_path):
                os.mkdir(har_path)
            profile.add_extension(os.path.join(self.cur_path, 'extensions/netExport-0.9b7.xpi'))
            profile.set_preference("extensions.firebug.DBG_NETEXPORT", True)
            profile.set_preference("extensions.firebug.netexport.alwaysEnableAutoExport", True)
            profile.set_preference("extensions.firebug.netexport.defaultLogDir", har_path)
            profile.set_preference("extensions.firebug.netexport.includeResponseBodies", True)
        return profile

    def open_virtual_display(self):
        self.display.start()

    def close_virtual_display(self):
        self.display.stop()

    def wrap_results(self, **kwargs):
        """
        Wrap returned http response into a well formatted dict
        :param kwargs: this dict param should contains following keys:
                           fd: file directory to
                           url: the test url fo the result
                           files_count: the number of files under har/ directory

        :return (dict): the results of all
        """

        if 'fd' not in kwargs \
                or 'url' not in kwargs \
                or 'files_count' not in kwargs:
            logging.error("Missing arguments in wrap_results function")
            return {}

        external = kwargs['external'] if 'external' in kwargs else None
        fd = kwargs['fd']
        url = kwargs['url']
        length = kwargs['files_count']

        results = {}

        files = []
        wait_time = 15
        host = self.divide_url(url)[0]
        time.sleep(0.5)

        # wait until the har file is generated
        while len(os.listdir(fd)) <= length + self.parsed:
            time.sleep(1)
            wait_time -= 1
            if wait_time == 0:
                logging.warning("%s waiting har file result timed out" % url)
                results['error'] = "wrap har file timeout"
                if external is not None:
                    external[url] = results
                return results
        time.sleep(1)

        # find all har files under har/ directory
        for fn in os.listdir(fd):
            if fn.endswith(".har") and host in fn:
                path = os.path.join(fd, fn)
                files.append((fn, os.stat(path).st_mtime))

        # sort all har files and parse the latest one
        files.sort(key=lambda x: x[1])
        if len(files) > 0:
            with open(fd + '/' + files[-1][0]) as f:
                raw_data = json.load(f)['log']['entries']
                results = [{} for i in range(0, len(raw_data))]
                for i in range(0, len(results)):

                    results[i]['request'] = {}
                    results[i]['request']['method'] = raw_data[i]['request']['method']
                    headers = {}
                    for header in raw_data[i]['request']['headers']:
                        headers[header['name']] = header['value']
                    results[i]['request']['headers'] = headers

                    results[i]['response'] = {}
                    results[i]['response']['status'] = raw_data[i]['response']['status']
                    results[i]['response']['reason'] = raw_data[i]['response']['statusText']
                    headers = {}
                    for header in raw_data[i]['response']['headers']:
                        headers[header['name']] = header['value']
                    results[i]['response']['headers'] = headers
                    results[i]['response']['redirect'] = raw_data[i]['response']['redirectURL']
                    results[i]['response']['body'] = raw_data[i]['response']['content']

            self.parsed += 1  # increment the number of parsed har files
        else:
            logging.warning("Cannot find har file for %s" % url)

        # save test result of this url to the external result object or 
        # return the result
        if external is not None:
            external[url] = results
        else:
            return results

    def divide_url(self, url):
        """
        divide url into host and path two parts
        """
        if 'https://' in url:
            host = url[8:].split('/')[0]
            path = url[8 + len(host):]
        elif 'http://' in url:
            host = url[7:].split('/')[0]
            path = url[7 + len(host):]
        else:
            host = url.split('/')[0]
            path = url[len(host):]
        return host, path

    def get(self, host, files_count, path="/", ssl=False, external=None):
        """
        Send get request to a url and wrap the results
        :param host (str): the host name of the url
        :param path (str): the path of the url (start with "/")
        :return (dict): the result of the test url
        """
        theme = "https" if ssl else "http"
        url = host + path
        http_url = theme + "://" + url

        result = {}
        try:
            capture_path = os.getcwd() + '/'

            har_file_path = capture_path + "har/"

            # fc.load_page(self.driver, http_url)
            fc.switch_tab(self.driver)
            self.load_page(http_url)

            print "driver get: " + http_url

            time.sleep(2)

            # if url[-1] == "/":
            #     f_name = url.split('/')[-2]
            # else:
            #     f_name = url.split('/')[-1]

            # fc.save_html(self.driver, f_name, os.path.join(capture_path, "htmls/"))
            # fc.save_screenshot(self.driver, f_name, os.path.join(capture_path, "screenshots/"))

            result = self.wrap_results(url=http_url, files_count=files_count, fd=har_file_path)

            if external is not None:
                external[http_url] = result

        except Exception as e:
            result['error'] = e.message
            print e

        return result

    def run_file(self, input_file, results):
        """
        use foctor_core library do get requests
        :param input_file: the file name of the list of test urls
                        format: 
                            1, www.facebook.com
                            2, www.google.com
                            ...
        :param results: the object to save the responses from server
        """

        capture_path = self.cur_path

        display_mode = 0  # 0 is virtural display(Xvfb mode)

        site_list = []
        file_name, file_contents = input_file
        result = {"file_name": file_name}
        file_metadata = {}
        file_comments = []
        run_start_time = time.time()
        index = 1

        csvreader = csv.reader(file_contents, delimiter=',', quotechar='"')
        for row in csvreader:
            """
                First few lines are expected to be comments in key: value
                format. The first line after that could be our column header
                row, starting with "url", and the rest are data rows.
                This is a sample input file we're trying to parse:

                # comment: Global List,,,,,
                # date: 03-17-2015,,,,,
                # version: 1,,,,,
                # description: This is the global list. Last updated in 2012.,,,,
                url,country,category,description,rationale,provider
                http://8thstreetlatinas.com,glo,PORN,,,PRIV
                http://abpr2.railfan.net,glo,MISC,Pictures of trains,,PRIV

                """

            # parse file comments, if it looks like "key : value",
            # parse it as a key-value pair. otherwise, just
            # store it as a raw comment.
            if row[0][0] == '#':
                row = row[0][1:].strip()
                if len(row.split(':')) > 1:
                    key, value = row.split(':', 1)
                    key = key.strip()
                    value = value.strip()
                    file_metadata[key] = value
                else:
                    file_comments.append(row)
                continue

            # detect the header row and store it
            # it is usually the first row and starts with "url,"
            if row[0].strip().lower() == "url":
                index_row = row
                continue

            url = row[0].strip()
            if url is None:
                continue

            meta = row[1:]
            site_list.append([index, url])
            index += 1

        driver, display = fc.do_crawl(sites=site_list, driver=self.driver, display=self.display,
                                      capture_path=capture_path, callback=self.wrap_results,
                                      external=results, fd=os.path.join(capture_path, "har/"),
                                      files_count=len(os.listdir(os.path.join(capture_path, "har/"))))
        fc.teardown_driver(driver, display, display_mode)

        driver.quit()  # quit driver will also clean up the tmp file under /tmp directory

    def run(self, input_files, url=None, verbose=0):
        """
        run the headless browser with given input
        if url given, the proc will only run hlb with given url and ignore input_list.
        :param url:
        :param input_files: the name of the file in "index url" format. i.e.
                1, www.facebook.com
                1, www.google.com
                ...
        :param verbose:
        :return:
        """
        if not url and not input_files:
            logging.warning("No input file")
            return {"error": "no inputs"}

        results = {}

        self.open_virtual_display()
        if verbose > 0:
            log_file = sys.stdout
        else:
            log_file = None

        # set up firefox driver 
        self.binary = FirefoxBinary(os.path.join(self.cur_path, 'firefox/firefox'), log_file=log_file)
        self.profile = self.setup_profile()
        self.driver = webdriver.Firefox(firefox_profile=self.profile, firefox_binary=self.binary, timeout=60)
        self.driver.set_page_load_timeout(60)

        isfile = False
        if url:
            host, path = self.divide_url(url)
            results[url] = self.get(host, path)
        else:
            isfile = True
            for input_file in input_files.items():
                logging.info("Testing input file %s..." % (input_file[0]))
                self.run_file(input_file, results)

        # foctor_core will quit the driver by itself so we only quit the driver when we don't use foctor core
        if not isfile:
            logging.info("Quit driver")
            self.driver.quit()
            self.close_virtual_display()

        logging.debug("Deleting har folder")
        shutil.rmtree(os.path.join(self.cur_path, 'har'))
        return results



from base64 import urlsafe_b64encode
import glob
import json
import logging
import os
import re
import requests
import time
import uuid

import centinel.utils as utils


logging.getLogger("requests").setLevel(logging.WARNING)


class User:
    def __init__(self, config):
        self.config = config
        self.verify = self.config['server']['verify']
        # check for login file
        if os.path.isfile(config['server']['login_file']):
            with open(config['server']['login_file']) as login_fh:
                login_details = json.load(login_fh)
                self.username = login_details.get('username')
                self.password = login_details.get('password')
                self.typeable_handle = login_details.get('typeable_handle')
                self.auth = (self.username, self.password)
        else:
            self.create_user()

    def request(self, slug):
        url = "%s/%s" % (self.config['server']['server_url'], slug)
        try:
            req = requests.get(url, auth=self.auth,
                               proxies=self.config['proxy']['proxy'],
                               verify=self.verify)
            req.raise_for_status()
            return req.json()
        except Exception as exp:
            logging.exception("Exception trying to make request - %s for URL: %s" %
                          (exp, url))
            raise exp

    @property
    def recommended_version(self):
        try:
            return int(self.request("version")["version"])
        except Exception as exp:
            logging.exception("Exception trying to get recommended version: %s " %
                          (exp))
            raise exp

    @property
    def experiments(self):
        try:
            return self.request("experiments")["experiments"]
        except Exception as exp:
            logging.exception("Error trying to get experiments: %s " % (exp))
            raise exp

    @property
    def input_files(self):
        try:
            return self.request("input_files")["inputs"]
        except Exception as exp:
            logging.exception("Error trying to get input files: %s " % (exp))
            raise exp

    @property
    def results(self):
        try:
            return self.request("results")
        except Exception as exp:
            logging.exception("Error trying to get results: %s " % (exp))
            raise exp

    @property
    def clients(self):
        try:
            return self.request("clients")
        except Exception as exp:
            logging.exception("Error trying to get clients: %s " % (exp))
            raise exp

    def submit_result(self, file_name):
        logging.info("Uploading result file - %s", file_name)

        with open(file_name) as result_file:
            files = {'result': result_file}
            url = "%s/%s" % (self.config['server']['server_url'], "results")
            timeout = self.config['server']['req_timeout']
            try:
                req = requests.post(url, files=files, auth=self.auth,
                                    proxies=self.config['proxy']['proxy'],
                                    timeout=timeout,
                                    verify=self.verify)
                req.raise_for_status()
                if ('delete_after_sync' in self.config['results'].keys()
                   and self.config['results']['delete_after_sync']):
                    os.remove(file_name)
            except Exception as exp:
                logging.exception("Error trying to submit result: %s" % exp)

    def sync_scheduler(self):
        """Download the scheduler.info file and perform a smart comparison
        with what we currently have so that we don't overwrite the
        last_run timestamp

        To do a smart comparison, we go over each entry in the
        server's scheduler file. If a scheduler entry is not present
        in the server copy, we delete it in the client copy and if the
        scheduler entry is present in the server copy, then we
        overwrite the frequency count in the client copy

        """
        # get the server scheduler.info file
        url = "%s/%s/%s" % (self.config['server']['server_url'],
                            "experiments", "scheduler.info")
        try:
            req = requests.get(url, proxies=self.config['proxy']['proxy'],
                               auth=self.auth,
                               verify=self.verify)
            req.raise_for_status()
        except Exception as exp:
            logging.exception("Error trying to download scheduler.info: %s" % exp)
            raise exp

        try:
            server_sched = json.loads(req.content)
        except Exception as exp:
            logging.exception("Error parsing server scheduler: %s" % exp)
            raise exp

        sched_filename = os.path.join(self.config['dirs']['experiments_dir'],
                                      'scheduler.info')
        if not os.path.exists(sched_filename):
            with open(sched_filename, 'w') as file_p:
                json.dump(server_sched, file_p, indent=2,
                      separators=(',', ': '))
            return

        client_sched = {}
        try:
            with open(sched_filename, 'r') as file_p:
                client_sched = json.load(file_p)
        except Exception as exp:
            client_sched = {}
            logging.exception("Error loading scheduler file: %s" % exp)
            logging.info("Making an empty scheduler")

        # delete any scheduled tasks as necessary
        #
        # Note: this looks ugly, but we can't modify dictionaries
        # while we iterate over them
        client_exp_keys = client_sched.keys()
        for exp in client_exp_keys:
            if exp not in server_sched:
                del client_sched[exp]
        # and update all the other frequencies
        for exp in server_sched:
            if exp in client_sched:
                client_sched[exp]['frequency'] = server_sched[exp]['frequency']
            else:
                client_sched[exp] = server_sched[exp]

        # write out the results
        with open(sched_filename, 'w') as file_p:
            json.dump(client_sched, file_p, indent=2,
                      separators=(',', ': '))

    def download_experiment(self, name):
        logging.info("Downloading experiment - %s", name)

        url = "%s/%s/%s" % (self.config['server']['server_url'],
                            "experiments", name)
        try:
            req = requests.get(url, proxies=self.config['proxy']['proxy'],
                               auth=self.auth,
                               verify=self.verify)
            req.raise_for_status()
        except Exception as exp:
            logging.exception("Error trying to download experiments: %s" % exp)
            raise exp

        name = "%s" % name
        with open(os.path.join(self.config['dirs']['experiments_dir'], name),
                  "w") as exp_fh:
            exp_fh.write(req.content)

    def download_input_file(self, name):
        logging.info("Downloading input data file - %s", name)

        url = "%s/%s/%s" % (self.config['server']['server_url'],
                            "input_files", name)
        try:
            req = requests.get(url, proxies=self.config['proxy']['proxy'],
                               auth=self.auth,
                               verify=self.verify)
            req.raise_for_status()
        except Exception as exp:
            logging.exception("Error trying to download experiments: %s" % exp)
            raise exp

        name = "%s" % name
        with open(os.path.join(self.config['dirs']['data_dir'], name),
                  "w") as exp_fh:
            exp_fh.write(req.content)

    def register(self, username, password):
        logging.info("Registering new user %s" % (username))

        url = "%s/%s" % (self.config['server']['server_url'], "register")
        payload = {'username': username, 'password': password,
                   'is_vpn': self.config['user'].get('is_vpn')}
        headers = {'content-type': 'application/json'}
        try:
            req = requests.post(url, data=json.dumps(payload),
                                proxies=self.config['proxy']['proxy'],
                                headers=headers,
                                verify=self.verify)
            req.raise_for_status()
            return req.json()
        except Exception as exp:
            logging.exception("Error trying to submit registration URL: %s " % exp)
            raise exp

    def set_country(self, country):
        url = "%s/%s/%s" % (self.config['server']['server_url'],
                            "set_country", country)
        try:
            req = requests.get(url,
                               auth=self.auth,
                               proxies=self.config['proxy']['proxy'],
                               verify=self.verify)
            req.raise_for_status()
            return req.json()
        except Exception as exp:
            logging.exception("Error trying to set country: %s " % exp)
            raise exp

    def set_ip(self, ip):
        url = "%s/%s/%s" % (self.config['server']['server_url'],
                            "set_ip", ip)
        try:
            req = requests.get(url,
                               auth=self.auth,
                               proxies=self.config['proxy']['proxy'],
                               verify=self.verify)
            req.raise_for_status()
            return req.json()
        except Exception as exp:
            logging.exception("Error trying to set ip: %s " % exp)
            raise exp

    def create_user(self):
        self.username = str(uuid.uuid4())
        self.password = os.urandom(64).encode('base-64')
        self.auth = (self.username, self.password)
        self.typeable_handle = None

        try:
            register_results = self.register(self.username, self.password)
            if 'typeable_handle' in register_results.keys():
                self.typeable_handle = register_results['typeable_handle']
            with open(self.config['server']['login_file'], "w") as login_fh:
                login_details = {'username': self.username,
                                 'password': self.password}
                if self.typeable_handle is not None:
                    login_details['typeable_handle'] = self.typeable_handle
                json.dump(login_details, login_fh, indent=2,
                      separators=(',', ': '))
        except Exception as exp:
            logging.exception("Unable to register: %s" % str(exp))
            raise exp

    def informed_consent(self):
        """Create a URL for the user to give their consent through"""
        if self.typeable_handle is None:
            consent_url = [self.config['server']['server_url'],
                           "/get_initial_consent?username="]
            consent_url.append(urlsafe_b64encode(self.username))
            consent_url.append("&password=")
            consent_url.append(urlsafe_b64encode(self.password))
        else:
            consent_url = [self.config['server']['server_url'],
                           "/consent/"]
            consent_url.append(self.typeable_handle)

        consent_url = "".join(consent_url)
        print "Please go to %s to give your consent." % (consent_url)
        return consent_url


def sync(config):
    logging.info("Starting sync with %s", config['server']['server_url'])

    start = time.time()
    try:
        user = User(config)
    except Exception, exp:
        logging.exception("Unable to create user: %s" % str(exp))
        return

    # send all results (.bz2)
    result_files = glob.glob(os.path.join(config['dirs']['results_dir'],
                                          '[!_]*.bz2'))

    # only upload pcaps if it is allowed
    if config['results']['upload_pcaps'] is False:
        for pcap_file in glob.glob(os.path.join(config['dirs']['results_dir'],
                                                '[!_]*.pcap.bz2')):
            if pcap_file in result_files:
                result_files.remove(pcap_file)

    for path in result_files:
        try:
            user.submit_result(path)
        except Exception, exp:
            if re.search("418", str(exp)) is not None:
                logging.error("You have not completed the informed consent "
                              "and will be unable to submit results or get "
                              "new experiments until you do.")
                user.informed_consent()
                return
            else:
                logging.error("Unable to send result file: %s" % str(exp))
            raise exp
        if time.time() - start > config['server']['total_timeout']:
            logging.error("Interaction with server took too long. Preempting")
            return

    # determine how to sync the experiment files
    # Note: we are not checking anything that starts with _
    client_exps = utils.hash_folder(config['dirs']['experiments_dir'],
                                    '[!_]*')
    try:
        server_exps = user.experiments
    except Exception as exp:
        if re.search("418", str(exp)) is not None:
            logging.error("You have not completed the informed consent and "
                          "will be unable to submit results or get new "
                          "experiments until you do.")
            user.informed_consent()
            return
        else:
            logging.error("Error collecting experiments: %s" % exp)
        raise exp
    if time.time() - start > config['server']['total_timeout']:
        logging.error("Interaction with server took too long. Preempting")
        return

    dload_exps, del_exps = utils.compute_files_to_download(client_exps,
                                                           server_exps)

    # delete the files that aren't on the server
    for exp_file in del_exps:
        filename = os.path.join(config['dirs']['experiments_dir'], exp_file)
        os.remove(filename)
    # get the files that have changed or we don't have
    for exp_file in dload_exps:
        try:
            if exp_file != "scheduler.info":
                user.download_experiment(exp_file)
            else:
                try:
                    user.sync_scheduler()
                except Exception as e:
                    logging.exception("Scheduler sync failed: %s", str(e))
        except Exception as e:
            logging.exception("Unable to download experiment file: %s", str(e))

        if time.time() - start > config['server']['total_timeout']:
            logging.error("Interaction with server took too long. Preempting")
            return

    # determine how to sync the input files
    client_inputs = utils.hash_folder(config['dirs']['data_dir'], '[!_]*')
    try:
        server_inputs = user.input_files
    except Exception as exp:
        logging.exception("Unable to retrive user inputs due to Exception: "
                      "%s. Preempting" % exp)
        return

    if time.time() - start > config['server']['total_timeout']:
        logging.error("Interaction with server took too long. Preempting")
        return

    dload_inputs, del_inputs = utils.compute_files_to_download(client_inputs,
                                                               server_inputs)

    # delete the files that aren't on the server
    for input_file in del_inputs:
        filename = os.path.join(config['dirs']['data_dir'], input_file)
        os.remove(filename)
    # get the files that have changed or we don't have
    for input_file in dload_inputs:
        try:
            user.download_input_file(input_file)
        except Exception, e:
            logging.exception("Unable to download input file %s", str(e))
        if time.time() - start > config['server']['total_timeout']:
            logging.error("Interaction with server took too long. Preempting")
            return

    logging.info("Finished sync with %s", config['server']['server_url'])


def set_vpn_info(config, ip=None, country=None):
    logging.info("Setting country as "
                 "%s and IP address as %s" % (country, ip))
    try:
        user = User(config)
    except Exception as exp:
        logging.exception("Unable to create user: %s" % str(exp))
        return False

    if country is not None:
        user.set_country(country)

    if ip is not None:
        user.set_ip(ip)


def get_meta(config, ip=''):
    url = "%s/%s/%s" % (config['server']['server_url'],
                        "meta", ip)
    try:
        req = requests.get(url,
                           proxies=config['proxy']['proxy'],
                           verify=config['server']['verify'],
                           timeout=10)
        req.raise_for_status()
        return req.json()
    except Exception as exp:
        logging.exception("Error trying to get metadata: %s " % exp)
        raise exp



import httplib2
import logging
import os
import shutil
import socket
import sys
import urllib
import zipfile


def unzip(source_filename, dest_dir):
    with zipfile.ZipFile(source_filename) as zf:
        zf.extractall(dest_dir)


def create_config_files(directory):
    """
    Initialize directory ready for vpn walker
    :param directory: the path where you want this to happen
    :return:
    """
    # Some constant strings
    config_zip_url = "http://www.ipvanish.com/software/configs/configs.zip"

    if not os.path.exists(directory):
        os.makedirs(directory)

    logging.info("Starting to download IPVanish config file zip")
    url_opener = urllib.URLopener()
    zip_path = os.path.join(directory, '../configs.zip')
    unzip_path = os.path.join(directory, '../unzipped')
    if not os.path.exists(unzip_path):
        os.makedirs(unzip_path)

    url_opener.retrieve(config_zip_url, zip_path)
    logging.info("Extracting zip file")
    unzip(zip_path, unzip_path)

    # remove zip file
    os.remove(zip_path)
    # copy ca and key to root path
    shutil.copyfile(os.path.join(unzip_path, 'ca.ipvanish.com.crt'),
                    os.path.join(directory, '../ca.ipvanish.com.crt'))

    # move all config files to /vpns

    server_country = {}
    for filename in os.listdir(unzip_path):
        if filename.endswith('.ovpn'):
            country = filename.split('-')[1]

            file_path = os.path.join(unzip_path, filename)
            lines = [line.rstrip('\n') for line in open(file_path)]

            # get ip address for this vpn
            ip = ""
            for line in lines:
                if line.startswith('remote'):
                    hostname = line.split(' ')[1]
                    ip = socket.gethostbyname(hostname)
                    break

            if len(ip) > 0:
                new_path = os.path.join(directory, ip + '.ovpn')
                shutil.copyfile(file_path, new_path)
                server_country[ip] = country
            else:
                logging.warn("Unable to resolve hostname and remove %s" % filename)
                os.remove(file_path)

    with open(os.path.join(directory, 'servers.txt'), 'w') as f:
        for ip in server_country:
            f.write('|'.join([ip, server_country[ip]]) + '\n')

    # remove extracted folder
    shutil.rmtree(unzip_path)

if __name__ == "__main__":
    if len(sys.argv) != 2:
        print "Usage {0} <directory to create VPNs in>".format(sys.argv[0])
        sys.exit(1)
    create_config_files(sys.argv[1])



from base64 import urlsafe_b64encode
import glob
import hashlib
import os.path

# Use this list to randomly choose a User-Agent string
user_agent_pool = [
    # Chrome
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/46.0.2490.80 Safari/537.36",
    # Firefox
    "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:40.0) Gecko/20100101 Firefox/40.1",
    # Internet Explorer
    "Mozilla/5.0 (compatible, MSIE 11, Windows NT 6.3; Trident/7.0; rv:11.0) like Gecko",
    # Maxthon
    "Mozilla/5.0 (Windows; U; Windows NT 6.0; en-US) AppleWebKit/533.1 "
    "(KHTML, like Gecko) Maxthon/3.0.8.2 Safari/533.1",
    # Opera
    "Opera/9.80 (X11; Linux i686; Ubuntu/14.10) Presto/2.12.388 Version/12.16",
    # Safari
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_3) AppleWebKit/537.75.14 "
    "(KHTML, like Gecko) Version/7.0.3 Safari/7046A194A",
    # MyIE2
    "Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 6.0; MyIE2; SLCC1; .NET CLR 2.0.50727; Media Center PC 5.0)"
]


def hash_folder(folder, regex='[!_]*'):
    """
    Get the md5 sum of each file in the folder and return to the user

    :param folder: the folder to compute the sums over
    :param regex: an expression to limit the files we match
    :return:

    Note: by default we will hash every file in the folder

    Note: we will not match anything that starts with an underscore

    """
    file_hashes = {}
    for path in glob.glob(os.path.join(folder, regex)):
        # exclude folders
        if not os.path.isfile(path):
            continue

        with open(path, 'r') as fileP:
            md5_hash = hashlib.md5(fileP.read()).digest()

        file_name = os.path.basename(path)
        file_hashes[file_name] = urlsafe_b64encode(md5_hash)
    return file_hashes


def compute_files_to_download(client_hashes, server_hashes):
    """
    Given a dictionary of file hashes from the client and the
    server, specify which files should be downloaded from the server

    :param client_hashes: a dictionary where the filenames are keys and the
                          values are md5 hashes as strings
    :param server_hashes: a dictionary where the filenames are keys and the
                          values are md5 hashes as strings
    :return: a list of 2 lists -> [to_dload, to_delete]
             to_dload- a list of filenames to get from the server
             to_delete- a list of filenames to delete from the folder

    Note: we will get a file from the server if a) it is not on the
    client or b) the md5 differs between the client and server

    Note: we will mark a file for deletion if it is not available on
    the server

    """
    to_dload, to_delete = [], []
    for filename in server_hashes:
        if filename not in client_hashes:
            to_dload.append(filename)
            continue
        if client_hashes[filename] != server_hashes[filename]:
            to_dload.append(filename)

    for filename in client_hashes:
        if filename not in server_hashes:
            to_delete.append(filename)

    return [to_dload, to_delete]



import os
from setuptools import setup, find_packages

DESCRIPTION = """\
Centinel is a tool used to detect network interference, \
online information controls, and Internet censorship."""

setup(
    name="centinel",
    version="0.1.5.7.1",
    author="ICLab Developers",
    author_email="info@iclab.org",
    description=DESCRIPTION,
    license="MIT",
    keywords="censorship information controls network interference",
    url="https://www.github.com/iclab/centinel",
    packages=["centinel", "centinel.primitives",
              "centinel.vpn"],
    install_requires=["argparse >= 1.2.1",
                      "dnspython >= 1.11.0",
                      "requests >= 2.9.1",
                      "trparse >= 0.2.1",
                      "pycurl >= 7.19.5",
                      "urllib3 >= 1.9.1",
                      "dnspython >= 1.12.0",
                      "BeautifulSoup >= 3.2.1",
                      "httplib2 >= 0.9.2",
                      "bs4 >= 0.0.1",
                      "geopy >= 1.11.0"],
    include_package_data=True,
    entry_points={
        'console_scripts': ['centinel=centinel.cli:run',
                            'centinel-vpn=centinel.vpn.cli:run']
    },
)



import pycurl
import re
from StringIO import StringIO


class ICHTTPConnection:

    def __init__(self, host='127.0.0.1', port=None, timeout=10):
        self.headers = {}
        self.body = None
        self.reason = None
        self.status = 0
        self.host = host
        self.port = port
        self.timeout = timeout

    def header_function(self, header_line):
        # HTTP standard specifies that headers are encoded in iso-8859-1.
        # On Python 2, decoding step can be skipped.
        # On Python 3, decoding step is required.
        # header_line = header_line.decode('iso-8859-1')

        # Header lines include the first status line (HTTP/1.x ...).
        # We are going to ignore all lines that don't have a colon in them.
        # This will botch headers that are split on multiple lines...
        if self.reason is None:
            reason = re.match('^HTTP/\d\.\d \d{3} (\w+)', header_line)
            if reason is not None:
                self.reason = reason.group(1)
        if ':' not in header_line:
            return

        # Break the header line into header name and value.
        name, value = header_line.split(':', 1)

        # Remove whitespace that may be present.
        # Header lines include the trailing newline, and there may be whitespace
        # around the colon.
        name = name.strip()
        value = value.strip()

        # Header names are case insensitive.
        # Lowercase name here.

        # Now we can actually record the header name and value.
        self.headers[name] = value

    def request(self, path="/", header=None, ssl=False, timeout=None):

        if timeout is None:
            timeout = self.timeout

        buf = StringIO()
        c = pycurl.Curl()

        if header:
            slist = []
            for key, value in header.iteritems():
                slist.append(key+": "+value)
            c.setopt(pycurl.HTTPHEADER, slist)

        c.setopt(pycurl.HEADERFUNCTION, self.header_function)
        c.setopt(pycurl.FOLLOWLOCATION, True)
        c.setopt(pycurl.WRITEDATA, buf)
        c.setopt(pycurl.TIMEOUT, timeout)
        c.setopt(pycurl.ENCODING, 'identity')
        c.setopt(pycurl.NOSIGNAL, 1)

        if ssl:
            if self.port is None:
                self.port = 443
            c.setopt(pycurl.URL, "https://"+self.host+":"+str(self.port)+path)
            c.setopt(pycurl.SSL_VERIFYPEER, 1)
            c.setopt(pycurl.SSL_VERIFYHOST, 2)
        else:
            if self.port is None:
                self.port = 80
            c.setopt(pycurl.URL,"http://"+self.host + ":" + str(self.port) + path)

        c.perform()

        self.status = c.getinfo(pycurl.RESPONSE_CODE)

        c.close()

        encoding = None
        if 'content-type' in self.headers:
            content_type = self.headers['content-type'].lower()
            match = re.search('charset=(\S+)', content_type)
            if match:
                encoding = match.group(1)
        if encoding is None:
            # Default encoding for HTML is iso-8859-1.
            # Other content types may have different default encoding,
            # or in case of binary data, may have no encoding at all.
            encoding = 'iso-8859-1'

        self.body = buf.getvalue().decode(encoding)



#!/usr/bin/python
__title__ = 'centinel'
__version__ = '0.1.5.7.1'

import centinel.backend
import centinel.client
import centinel.command
import centinel.config
import centinel.cli
import centinel.daemonize
import centinel.utils



#!/usr/bin/python
#
# daemonize.py: functionality to make centinel run in the background

import os
import shutil
import stat
import tempfile


def create_script_for_location(content, destination):
    """Create a script with the given content, mv it to the
    destination, and make it executable

    Parameters:
    content- the content to put in the script
    destination- the directory to copy to

    Note: due to constraints on os.rename, destination must be an
    absolute path to a file, not just a directory

    """
    temp = tempfile.NamedTemporaryFile(mode='w', delete=False)
    temp.write(content)
    temp.close()
    shutil.move(temp.name, destination)
    cur_perms = os.stat(destination).st_mode
    set_perms = cur_perms | stat.S_IXOTH | stat.S_IXGRP | stat.S_IXUSR
    os.chmod(destination, set_perms)


def daemonize(package, bin_loc, user):
    """Create crontab entries to run centinel every hour and
    autoupdate every day

    Parameters:

    package- name of the currently installed package (will be used for
        autoupdate). If this parameter is None, the autoupdater will
        not be used

    bin_loc- location of the centinel binary/script.

    Note: this works by creating temporary files, adding the content
    of the cron scripts to these temporary files, moving these files
    into the appropriate cron folders, and making these scripts
    executable

    Note: if the script already exists, this will delete it

    """

    path = "/etc/cron.hourly/centinel-" + user

    if user != "root":
        # create a script to run centinel every hour as the current user
        hourly = "".join(["#!/bin/bash\n",
                          "# cron job for centinel\n",
                          "su ", user, " -c '", bin_loc, " --sync'\n",
                          "su ", user, " -c '", bin_loc, "'\n",
                          "su ", user, " -c '", bin_loc, " --sync'\n"])
    else:
        # create a script to run centinel every hour as root
        hourly = "".join(["#!/bin/bash\n",
                          "# cron job for centinel\n",
                          bin_loc, " --sync\n",
                          bin_loc, "\n",
                          bin_loc, " --sync\n"])

    create_script_for_location(hourly, path)

    # create a script to get the client to autoupdate every day
    if package is None:
        return
    updater = "".join(["#!/bin/bash\n",
                      "# autoupdater for centinel\n"
                      "sudo pip install --upgrade ", package, "\n"])
    create_script_for_location(updater, "/etc/cron.daily/centinel-autoupdate")
    print "Successfully created cron jobs for user " + user




# Ben Jones bjones99@gatech.edu
# Georgia Tech Fall 2014
#
# tcpdump.py: interface to tcpdump to stop and start captures and do
# second passes over existing pcaps

from base64 import b64encode
import logging
import os
import tempfile


# local imports
import centinel
from centinel import command


class Tcpdump():
    """Class to interface between tcpdump and Python"""

    def __init__(self, filename=None, pcap_args=None):
        if filename is None:
            temp_file = tempfile.NamedTemporaryFile(mode='w', delete=False)
            temp_file.close()
            filename = temp_file.name
        self.filename = filename
        # don't change this to default value because it is a mutable
        # type and whatever you change it to will become the new
        # default value until the interpreter is restarted
        if pcap_args is None:
            # use the centinel configured tcpdump options if available
            # (if not specified by the user, this will be -i any, so
            # the same as below
            if 'tcpdump_params' in centinel.conf['experiments']:
                pcap_args = centinel.conf['experiments']['tcpdump_params']
            # for backwards compatability, ensure that we give some
            # pcap args for what to capture
            else:
                pcap_args = ["-i", "any"]
                logging.warning("Global config not available, so falling "
                                "back on -i any pcap args")
        self.pcap_args = pcap_args

    def start(self):
        cmd = ['sudo', 'tcpdump', '-w', self.filename]
        cmd.extend(self.pcap_args)
        self.caller = command.Command(cmd, _tcpdump_callback)
        self.caller.start()

    def stop(self):
        if self.caller is not None:
            self.caller.stop()

    def post_processing(self, out_filter, out_file=None):
        if out_file is None:
            temp_file = tempfile.NamedTemporaryFile(mode='w', delete=False)
            temp_file.close()
            out_file = temp_file.name
        cmd = ['tcpdump', '-r', self.filename, '-w', out_file]
        caller = command.Command(cmd, _tcpdump_callback)
        caller.start()

    def b64_output(self):
        with open(self.filename, 'r') as file_p:
            return b64encode(file_p.read())

    def pcap(self):
        with open(self.filename, 'r') as file_p:
            return file_p.read()

    def pcap_filename(self):
        return self.filename

    def delete(self):
        os.remove(self.filename)


def _tcpdump_callback(self, line, kill_switch):
    """Callback function to handle tcpdump"""

    line = line.lower()
    if ("listening" in line) or ("reading" in line):
        self.started = True
    if ("no suitable device" in line):
        self.error = True
        self.kill_switch()
    if "by kernel" in line:
        self.stopped = True



#!/usr/bin/python
# hma.py: collection of functions to create vpn config files from the
# Hide My Ass VPN service

import os
import requests
import sys


def create_config_files(directory):
    """Create all available VPN configuration files in the given directory

    Note: I am basically just following along with what their script
    client does

    """
    # get the config file template
    template_url = ("https://securenetconnection.com/vpnconfig/"
                    "openvpn-template.ovpn")
    resp = requests.get(template_url)
    resp.raise_for_status()
    template = resp.content

    # get the available servers and create a config file for each server
    server_url = ("https://securenetconnection.com/vpnconfig/"
                  "servers-cli.php")
    resp = requests.get(server_url)
    resp.raise_for_status()
    servers = resp.content.split("\n")

    if not os.path.exists(directory):
        os.makedirs(directory)
    with open(os.path.join(directory, "servers.txt"), 'w') as f:
        f.write(resp.content)

    for server_line in servers:
        if server_line.strip() == "":
            continue
        server_line = server_line.split("|")
        try:
            ip, desc, country, udp_sup, tcp_sup = server_line
        except ValueError:
            ip, desc, country, udp_sup, tcp_sup, no_rand = server_line
        with open(os.path.join(directory, ip + ".ovpn"), 'w') as file_o:
            file_o.write(template)
            # create tcp if available, else udp
            tcp_sup = tcp_sup.strip()
            if tcp_sup:
                port, proto = 443, "tcp"
            else:
                port, proto = 53, "udp"
            file_o.write("remote {0} {1}\n".format(ip, port))
            file_o.write("proto {0}\n".format(proto))
            # add automatic dns server update
            file_o.write("up /etc/openvpn/update-resolv-conf\n")
            file_o.write("down /etc/openvpn/update-resolv-conf\n")


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print "Usage {0} <directory to create VPNs in>".format(sys.argv[0])
        sys.exit(1)
    create_config_files(sys.argv[1])
